from typing import Any, Callable

from frozendict import frozendict

from bgpsimulator.shared import ASNGroups, CycleError

from .base_as import AS


class ASGraphUtils:
    """Utility functions for ASGraph"""

    @staticmethod
    def add_extra_setup(
        as_graph_json: dict[str, Any],
        additional_asn_group_filters: frozendict[
            str, Callable[[dict[int, AS]], frozenset[int]]
        ] = frozendict(),
    ) -> None:
        """Adds cycles, provider cone, asn_groups, and propagation ranks
        to the AS graph
        """

        if not as_graph_json.get("extra_setup_complete", False):
            # Conver to ints when pulling from JSON
            as_graph_json["ases"] = {
                int(asn): info for asn, info in as_graph_json["ases"].items()
            }
            ASGraphUtils.check_for_cycles(as_graph_json)
            ASGraphUtils.add_provider_cone_asns(as_graph_json)
            ASGraphUtils.assign_as_propagation_rank(as_graph_json)
            ASGraphUtils.assign_as_graph_propagation_ranks(as_graph_json)
            ASGraphUtils.add_asn_groups(as_graph_json, additional_asn_group_filters)
            as_graph_json["extra_setup_complete"] = True

    ###############
    # Cycle funcs #
    ###############

    @staticmethod
    def check_for_cycles(as_graph_json: dict[str, Any]) -> None:
        """Checks for cycles in the AS graph"""

        # Apply cycle detection to each node in the graph
        for key in ("provider_asns", "customer_asns"):
            visited: set[int] = set()  # To track nodes that have been fully processed
            rec_stack: set[int] = (
                set()
            )  # Tracks current recursion stack (for cycle detection)

            for asn, as_info in as_graph_json["ases"].items():
                if asn not in visited:
                    ASGraphUtils._validate_no_cycles_helper(
                        asn, as_info, as_graph_json, visited, rec_stack, key
                    )
        as_graph_json["cycles_detected"] = False

    @staticmethod
    def _validate_no_cycles_helper(
        asn: int,
        as_info: dict[str, Any],
        as_graph_json: dict[str, Any],
        visited: set[int],
        rec_stack: set[int],
        key: str,
    ) -> None:
        """Helper function to detect cycles using DFS"""

        if asn not in visited:
            visited.add(asn)
            rec_stack.add(asn)

            # Visit all the providers (similar to graph neighbors) recursively
            for neighbor_asn in as_info.get(key, []):
                if neighbor_asn not in visited:
                    ASGraphUtils._validate_no_cycles_helper(
                        neighbor_asn,
                        as_graph_json["ases"][neighbor_asn],
                        as_graph_json,
                        visited,
                        rec_stack,
                        key,
                    )
                elif neighbor_asn in rec_stack:
                    raise CycleError(f"Cycle detected in {key} for AS {asn}")

        rec_stack.remove(asn)

    #################
    # Provider cone #
    #################

    @staticmethod
    def add_provider_cone_asns(as_graph_json: dict[str, Any]) -> None:
        """Adds provider cone ASNs to the AS graph"""

        cone_dict: dict[int, set[int]] = {}
        for _asn, as_info in as_graph_json["ases"].items():
            provider_cone: set[int] = ASGraphUtils._get_cone_helper(
                as_info, cone_dict, as_graph_json, "provider_asns"
            )
            as_info["provider_cone_asns"] = list(provider_cone)

    @staticmethod
    def _get_cone_helper(
        as_info: dict[str, Any],
        cone_dict: dict[int, set[int]],
        as_graph_json: dict[str, Any],
        rel_key: str,
    ) -> set[int]:
        """Recursively determines the cone of an AS"""

        as_asn = as_info["asn"]
        if as_asn in cone_dict:
            return cone_dict[as_asn]
        else:
            cone_dict[as_asn] = set()
            for neighbor_asn in as_info.get(rel_key, []):
                cone_dict[as_asn].add(neighbor_asn)
                if neighbor_asn not in cone_dict:
                    ASGraphUtils._get_cone_helper(
                        as_graph_json["ases"][neighbor_asn],
                        cone_dict,
                        as_graph_json,
                        rel_key,
                    )
                cone_dict[as_asn].update(cone_dict[neighbor_asn])
        return cone_dict[as_asn]

    ##########################
    # Propagation rank funcs #
    ##########################

    @staticmethod
    def assign_as_propagation_rank(as_graph_json: dict[str, Any]) -> None:
        """Adds propagation rank from the leafs to the input clique"""

        for as_info in as_graph_json["ases"].values():
            # Always set this to None since you can't trust this value
            as_info["propagation_rank"] = None
        for as_info in as_graph_json["ases"].values():
            ASGraphUtils._assign_ranks_helper(as_info, 0, as_graph_json)

    @staticmethod
    def _assign_ranks_helper(
        as_info: dict[str, Any], rank: int, as_graph_json: dict[str, Any]
    ) -> None:
        """Assigns ranks to all ases in customer/provider chain recursively"""

        if (
            as_info.get("propagation_rank") is None
            or as_info["propagation_rank"] < rank
        ):
            as_info["propagation_rank"] = rank
            # Only update it's providers if it's rank becomes higher
            # This avoids a double for loop of writes
            for provider_asn in as_info.get("provider_asns", []):
                ASGraphUtils._assign_ranks_helper(
                    as_graph_json["ases"][provider_asn], rank + 1, as_graph_json
                )

    @staticmethod
    def assign_as_graph_propagation_ranks(
        as_graph_json: dict[str, Any],
    ) -> None:
        """Orders ASes by rank"""

        max_rank: int = max(
            x["propagation_rank"] for x in as_graph_json["ases"].values()
        )
        # Create a list of empty lists
        # Ignore types here for speed purposes
        ranks: list[list[int]] = [list() for _ in range(max_rank + 1)]
        # Append the ASes into their proper rank
        for asn, as_info in as_graph_json["ases"].items():
            ranks[as_info["propagation_rank"]].append(asn)

        # Create tuple ranks
        as_graph_json["propagation_rank_asns"] = [sorted(rank) for rank in ranks]

    ####################
    # ASN groups funcs #
    ####################

    @staticmethod
    def add_asn_groups(
        as_graph_json: dict[str, Any],
        additional_asn_group_filters: frozendict[
            str, Callable[[dict[int, AS]], frozenset[int]]
        ],
    ) -> None:
        """Gets ASN groups. Used for choosing attackers from stubs, adopters, etc."""

        asn_to_as: dict[int, AS] = {
            asn: AS.from_json(as_info) for asn, as_info in as_graph_json["ases"].items()
        }

        asn_group_filters: dict[str, Callable[[dict[int, AS]], frozenset[int]]] = dict(
            **ASGraphUtils.get_default_as_group_filters(),
            **additional_asn_group_filters,
        )

        asn_groups: frozendict[str, frozenset[int]] = frozendict(
            {
                asn_group_key: filter_func(asn_to_as)
                for asn_group_key, filter_func in asn_group_filters.items()
            }
        )

        as_graph_json["asn_groups"] = {
            k: list(asn_group) for k, asn_group in asn_groups.items()
        }

    @staticmethod
    def get_default_as_group_filters() -> dict[
        str, Callable[[dict[int, AS]], frozenset[int]]
    ]:
        """Returns the default filter functions for AS groups"""

        def ixp_filter(asn_to_as: dict[int, AS]) -> frozenset[int]:
            return frozenset(asn for asn, as_ in asn_to_as.items() if as_.ixp)

        def stub_no_ixp_filter(asn_to_as: dict[int, AS]) -> frozenset[int]:
            return frozenset(
                asn for asn, as_ in asn_to_as.items() if as_.stub and not as_.ixp
            )

        def multihomed_no_ixp_filter(asn_to_as: dict[int, AS]) -> frozenset[int]:
            return frozenset(
                asn for asn, as_ in asn_to_as.items() if as_.multihomed and not as_.ixp
            )

        def stubs_or_multihomed_no_ixp_filter(
            asn_to_as: dict[int, AS],
        ) -> frozenset[int]:
            return frozenset(
                asn
                for asn, as_ in asn_to_as.items()
                if (as_.stub or as_.multihomed) and not as_.ixp
            )

        def tier_1_no_ixp_filter(asn_to_as: dict[int, AS]) -> frozenset[int]:
            return frozenset(
                asn for asn, as_ in asn_to_as.items() if as_.tier_1 and not as_.ixp
            )

        def etc_no_ixp_filter(asn_to_as: dict[int, AS]) -> frozenset[int]:
            return frozenset(
                asn
                for asn, as_ in asn_to_as.items()
                if not (as_.stub or as_.multihomed or as_.tier_1 or as_.ixp)
            )

        def transit_no_ixp_filter(asn_to_as: dict[int, AS]) -> frozenset[int]:
            return frozenset(
                asn for asn, as_ in asn_to_as.items() if as_.transit and not as_.ixp
            )

        def all_no_ixp_filter(asn_to_as: dict[int, AS]) -> frozenset[int]:
            return frozenset(asn_to_as.keys())

        return {
            ASNGroups.IXPS: ixp_filter,
            ASNGroups.STUBS: stub_no_ixp_filter,
            ASNGroups.MULTIHOMED: multihomed_no_ixp_filter,
            ASNGroups.STUBS_OR_MH: stubs_or_multihomed_no_ixp_filter,
            ASNGroups.TIER_1: tier_1_no_ixp_filter,
            ASNGroups.ETC: etc_no_ixp_filter,
            ASNGroups.TRANSIT: transit_no_ixp_filter,
            ASNGroups.ALL_WOUT_IXPS: all_no_ixp_filter,
        }



from typing import Any, ClassVar

from bgpsimulator.as_graphs.as_graph.as_graph import ASGraph
from bgpsimulator.simulation_framework.scenarios import ScenarioConfig


class EngineRunConfig:
    """Gemerates a JSON config for a single engine run

    Useful for tests and diagrams
    """

    _used_names: ClassVar[set[str]] = set()

    def __init__(
        self,
        name: str,
        scenario_config: ScenarioConfig,
        as_graph: ASGraph,
        diagram_desc: str = "",
        text: str = "",
        lab_text: str = "",
        diagram_ranks: list[list[int]] | None = None,
        prevent_naming_duplicates: bool = True,
    ):
        self.name = name
        self.prevent_naming_duplicates: bool = prevent_naming_duplicates

        # Useful for pytest, but turned off for the website
        if self.name in EngineRunConfig._used_names and prevent_naming_duplicates:
            raise ValueError(f"Name {self.name} already used")
        EngineRunConfig._used_names.add(self.name)
        self.diagram_desc = diagram_desc
        # Displayed in the website giant text box
        self.text = text
        self.lab_text = lab_text
        self.scenario_config = scenario_config
        self.as_graph = as_graph
        self.diagram_ranks = diagram_ranks or []

    def __eq__(self, other):
        if isinstance(other, EngineRunConfig):
            return (
                self.scenario_config == other.scenario_config
                and self.as_graph == other.as_graph
            )
        else:
            return NotImplemented

    def to_json(self) -> dict[str, Any]:
        """Converts the engine run config to a JSON object"""

        return {
            "name": self.name,
            "diagram_desc": self.diagram_desc,
            "text": self.text,
            "lab_text": self.lab_text,
            "scenario_config": self.scenario_config.to_json(),
            "as_graph": self.as_graph.to_json(),
            "diagram_ranks": self.diagram_ranks,
            "prevent_naming_duplicates": self.prevent_naming_duplicates,
        }

    @classmethod
    def from_json(cls, json_obj: dict[str, Any]) -> "EngineRunConfig":
        """Converts a JSON object to an engine run config"""
        return cls(
            name=json_obj.get("name", ""),
            diagram_desc=json_obj.get("diagram_desc", ""),
            text=json_obj.get("text", ""),
            lab_text=json_obj.get("lab_text", ""),
            scenario_config=ScenarioConfig.from_json(json_obj["scenario_config"]),
            as_graph=ASGraph.from_json(json_obj["as_graph"]),
            diagram_ranks=json_obj.get("diagram_ranks"),
            prevent_naming_duplicates=json_obj.get("prevent_naming_duplicates", False),
        )



from functools import cached_property
from typing import Any

from bgpsimulator.route_validator import ROA
from bgpsimulator.shared import ASNGroups, IPAddr, Settings
from bgpsimulator.simulation_engine import Announcement as Ann

from .scenario import Scenario


class ScenarioConfig:
    """Config reused across trials to set up a scenario/attack"""

    def __eq__(self, other):
        if isinstance(other, ScenarioConfig):
            return self.to_json() == other.to_json()
        else:
            return NotImplemented

    def __init__(
        self,
        label: str,
        ScenarioCls: type["Scenario"],
        propagation_rounds: int | None = None,
        attacker_settings: dict[Settings, bool] | None = None,
        legitimate_origin_settings: dict[Settings, bool] | None = None,
        override_adoption_settings: dict[int, dict[Settings, bool]] | None = None,
        override_base_settings: dict[int, dict[Settings, bool]] | None = None,
        default_adoption_settings: dict[Settings, bool] | None = None,
        default_base_settings: dict[Settings, bool] | None = None,
        # NOTE: we don't want behavior where attackers and victims are
        # randomly added... especially for the website. It was never
        # proper behavior though, so we default to 0.
        num_attackers: int = 0,
        num_legitimate_origins: int = 0,
        attacker_asn_group: str = ASNGroups.STUBS_OR_MH.value,
        legitimate_origin_asn_group: str = ASNGroups.STUBS_OR_MH.value,
        adoption_asn_groups: list[str] | None = None,
        override_attacker_asns: set[int] | None = None,
        override_legitimate_origin_asns: set[int] | None = None,
        override_adopting_asns: set[int] | None = None,
        override_seed_asn_ann_dict: dict[int, list[Ann]] | None = None,
        override_roas: list[ROA] | None = None,
        override_dest_ip_addr: IPAddr | None = None,
    ):
        # Label used for graphing, typically name it after the adopting policy
        self.label: str = label
        self.ScenarioCls: type[Scenario] = ScenarioCls

        ###########################
        # Routing Policy Settings #
        ###########################

        # When determining if an AS is using a setting, the following order is used:
        # 1. attacker_settings or legitimate_origin_settings (if AS is an attacker or
        #    legitimate_origin)
        # 2. override_adoption_settings (if set)
        # 3. override_base_settings
        # 4. default_adoption_settings
        # 5. default_base_settings

        # 1a. This will update the base routing policy settings for the attacker ASes
        self.attacker_settings: dict[Settings, bool] = (
            self.update_supersets(attacker_settings) or dict()
        )
        # 1v. This will update the base routing policy settings for the
        #     legitimate_origin ASes
        self.legitimate_origin_settings: dict[Settings, bool] = (
            self.update_supersets(legitimate_origin_settings) or dict()
        )
        # 2. This will completely override the default adopt routing policy settings
        self.override_adoption_settings: dict[int, dict[Settings, bool]] = dict()
        if override_adoption_settings:
            self.override_adoption_setting = {
                asn: self.update_supersets(settings_dict)
                for asn, settings_dict in override_adoption_settings.items()
            }
        # 3. This will completely override the default base routing policy settings
        self.override_base_settings: dict[int, dict[Settings, bool]] = dict()
        if override_base_settings:
            self.override_base_settings = {
                asn: self.update_supersets(settings_dict)
                for asn, settings_dict in override_base_settings.items()
            }
        # 4. This will update the base routing policy settings for the adopting ASes
        self.default_adoption_settings: dict[Settings, bool] = (
            self.update_supersets(default_adoption_settings) or dict()
        )
        # 5. Base routing policy settings that will be applied to all ASes
        self.default_base_settings: dict[Settings, bool] = (
            self.update_supersets(default_base_settings) or dict()
        )

        # Attackers are randomly selected from this ASN group
        self.attacker_asn_group: str = attacker_asn_group
        # Victims are randomly selected from this ASN group
        self.legitimate_origin_asn_group: str = legitimate_origin_asn_group
        # Adoption is equal across these ASN groups
        self.adoption_asn_groups: list[str] = adoption_asn_groups or [
            ASNGroups.STUBS_OR_MH.value,
            ASNGroups.ETC.value,
            ASNGroups.TIER_1.value,
        ]

        # Number of attackers/legitimate_origins/adopting ASes
        self.num_attackers: int = num_attackers
        self.num_legitimate_origins: int = num_legitimate_origins

        # Forces the attackers/legitimate_origins/adopting ASes to be a specific set
        # of ASes rather than random
        self.override_attacker_asns: set[int] | None = override_attacker_asns
        if self.override_attacker_asns is not None:
            self.num_attackers = len(self.override_attacker_asns)
        self.override_legitimate_origin_asns: set[int] | None = (
            override_legitimate_origin_asns
        )
        if self.override_legitimate_origin_asns is not None:
            self.num_legitimate_origins = len(self.override_legitimate_origin_asns)
        self.override_adopting_asns: set[int] | None = override_adopting_asns
        # Forces the announcements/roas to be a specific set of announcements/roas
        # rather than generated dynamically based on attackers/legitimate_origins

        self.override_seed_asn_ann_dict: dict[int, list[Ann]] | None = (
            override_seed_asn_ann_dict
        )
        self.override_roas: list[ROA] | None = override_roas
        # Every AS will attempt to send a packet to this IP address post propagation
        # This is used for the ASGraphAnalyzer to determine the outcome of a packet
        self.override_dest_ip_addr: IPAddr | None = override_dest_ip_addr

        # Below this point 99% of devs will not need to touch
        self.propagation_rounds = propagation_rounds
        if self.propagation_rounds is None:
            # BGP-iSec needs this.
            if any(
                x in self.all_used_settings
                for x in [Settings.BGP_I_SEC, Settings.BGP_I_SEC_TRANSITIVE]
            ):
                from bgpsimulator.simulation_framework.scenarios.shortest_path_prefix_hijack import (  # noqa: E501
                    ShortestPathPrefixHijack,
                )

                if issubclass(self.ScenarioCls, ShortestPathPrefixHijack):
                    # ShortestPathPrefixHijack needs 2 propagation rounds
                    self.propagation_rounds: int = 2
                else:
                    self.propagation_rounds = self.ScenarioCls.min_propagation_rounds
            else:
                self.propagation_rounds = self.ScenarioCls.min_propagation_rounds
        if self.ScenarioCls.min_propagation_rounds > self.propagation_rounds:
            raise ValueError(
                f"{self.ScenarioCls.__name__} requires a minimum of "
                f"{self.ScenarioCls.min_propagation_rounds} propagation rounds "
                f"but this scenario_config has only {self.propagation_rounds} "
                "propagation rounds"
            )
        self.modify_for_withdrawals()

    def modify_for_withdrawals(self) -> bool:
        """Modifies the required settings for withdrawals"""

        # NOTE: ROST also deals with withdrawals, must be in this list
        requires_bgp_full = (
            Settings.LEAKER,
            Settings.ANNOUNCE_THEN_WITHDRAW,
            Settings.ROST,
        )
        if any(setting in self.all_used_settings for setting in requires_bgp_full):
            self.default_base_settings[Settings.BGP_FULL] = True
            if self.propagation_rounds == 1:
                self.propagation_rounds = 2

    @staticmethod
    def update_supersets(
        policy_settings: dict[Settings, bool] | None,
    ) -> dict[Settings, bool]:
        """Updates the supersets of a policy setting"""

        if policy_settings is None:
            return dict()

        new_settings = policy_settings.copy()
        if policy_settings.get(Settings.BGP_I_SEC, False):
            new_settings[Settings.BGP_I_SEC_TRANSITIVE] = True
            new_settings[Settings.PROVIDER_CONE_ID] = True
            new_settings[Settings.ONLY_TO_CUSTOMERS] = True
        if policy_settings.get(Settings.PATH_END, False):
            new_settings[Settings.ROV] = True
        if any(
            policy_settings.get(x)
            for x in [Settings.ROVPP_V2_LITE, Settings.ROVPP_V2I_LITE]
        ):
            new_settings[Settings.ROVPP_V1_LITE] = True
            new_settings[Settings.ROV] = True
        if policy_settings.get(Settings.ROVPP_V1_LITE, False):
            new_settings[Settings.ROV] = True
        if any(policy_settings.get(x) for x in [Settings.ASRA, Settings.ASPA_W_N]):
            new_settings[Settings.ASPA] = True
        if any(policy_settings.get(x) for x in [Settings.ASPAPP]):
            new_settings[Settings.ASPA] = True
            new_settings[Settings.ASRA] = True
            new_settings[Settings.PROVIDER_CONE_ID] = True
        return new_settings

    @cached_property
    def all_used_settings(self) -> frozenset[Settings]:
        """Returns all the settings that are used in the scenario config"""
        used_settings = set()
        for settings_dict in [
            self.attacker_settings,
            self.legitimate_origin_settings,
            self.default_adoption_settings,
            self.default_base_settings,
        ]:
            # Mypy can't detect this type is {Settings: bool} for some reason
            for setting, bool_val in settings_dict.items():
                if bool_val:
                    used_settings.add(setting)
        for asn_to_settings_dict in (
            self.override_adoption_settings,
            self.override_base_settings,
        ):
            for settings_dict in asn_to_settings_dict.values():
                for setting, bool_val in settings_dict.items():
                    if bool_val:
                        used_settings.add(setting)
        return frozenset(used_settings)

    ##############
    # JSON Funcs #
    ##############

    def to_json(self) -> dict[str, Any]:
        """Converts the scenario config to a JSON object"""
        # Only include fields that are constructor parameters
        vals = {
            "label": self.label,
            "ScenarioCls": self.ScenarioCls.__name__,
            "propagation_rounds": self.propagation_rounds,
            "attacker_settings": self.attacker_settings,
            "legitimate_origin_settings": self.legitimate_origin_settings,
            "override_adoption_settings": self.override_adoption_settings,
            "override_base_settings": self.override_base_settings,
            "default_adoption_settings": self.default_adoption_settings,
            "default_base_settings": self.default_base_settings,
            "num_attackers": self.num_attackers,
            "num_legitimate_origins": self.num_legitimate_origins,
            "attacker_asn_group": self.attacker_asn_group,
            "legitimate_origin_asn_group": self.legitimate_origin_asn_group,
            "adoption_asn_groups": self.adoption_asn_groups,
            "override_attacker_asns": self.override_attacker_asns,
            "override_legitimate_origin_asns": self.override_legitimate_origin_asns,
            "override_adopting_asns": self.override_adopting_asns,
            "override_seed_asn_ann_dict": self.override_seed_asn_ann_dict,
            "override_roas": self.override_roas,
            "override_dest_ip_addr": self.override_dest_ip_addr,
        }

        # Handle override_seed_asn_ann_dict
        if vals.get("override_seed_asn_ann_dict") is not None:
            vals["override_seed_asn_ann_dict"] = {
                str(asn): [ann.to_json() for ann in anns]
                for asn, anns in vals["override_seed_asn_ann_dict"].items()  # type: ignore
            }

        # Handle override_roas
        if vals.get("override_roas") is not None:
            vals["override_roas"] = [roa.to_json() for roa in vals["override_roas"]]  # type: ignore

        # Handle override_dest_ip_addr
        if vals.get("override_dest_ip_addr") is not None:
            vals["override_dest_ip_addr"] = str(vals["override_dest_ip_addr"])

        # Convert sets to lists for JSON serialization
        if vals.get("override_attacker_asns") is not None:
            vals["override_attacker_asns"] = list(vals["override_attacker_asns"])  # type: ignore
        if vals.get("override_legitimate_origin_asns") is not None:
            vals["override_legitimate_origin_asns"] = list(  # type: ignore
                vals["override_legitimate_origin_asns"]
            )
        if vals.get("override_adopting_asns") is not None:
            vals["override_adopting_asns"] = list(vals["override_adopting_asns"])  # type: ignore

        return vals

    @classmethod
    def from_json(cls, json_obj: dict[str, Any]) -> "ScenarioConfig":
        """Converts a JSON object to a scenario config"""
        vals = json_obj.copy()
        vals["ScenarioCls"] = Scenario.name_to_cls_dict[vals["ScenarioCls"]]

        # Handle override_seed_asn_ann_dict
        if vals.get("override_seed_asn_ann_dict") is not None:
            vals["override_seed_asn_ann_dict"] = {
                int(asn): [Ann.from_json(ann) for ann in anns]
                for asn, anns in vals["override_seed_asn_ann_dict"].items()
            }

        fields_to_convert_keys_to_ints = (
            "override_adoption_settings",
            "override_base_settings",
        )
        for field in fields_to_convert_keys_to_ints:
            if vals.get(field):
                vals[field] = {int(k): v for k, v in vals[field].items()}

        # Handle override_roas
        if vals.get("override_roas") is not None:
            from bgpsimulator.route_validator import ROA

            vals["override_roas"] = [
                ROA.from_json(roa) for roa in vals["override_roas"]
            ]

        # Handle override_dest_ip_addr
        if vals.get("override_dest_ip_addr") is not None:
            from bgpsimulator.shared import IPAddr

            vals["override_dest_ip_addr"] = IPAddr(vals["override_dest_ip_addr"])

        # Convert lists back to sets
        if vals.get("override_attacker_asns") is not None:
            vals["override_attacker_asns"] = set(vals["override_attacker_asns"])
        if vals.get("override_legitimate_origin_asns") is not None:
            vals["override_legitimate_origin_asns"] = set(
                vals["override_legitimate_origin_asns"]
            )
        if vals.get("override_adopting_asns") is not None:
            vals["override_adopting_asns"] = set(vals["override_adopting_asns"])

        return cls(**vals)



from collections import defaultdict
from typing import TYPE_CHECKING

from bgpsimulator.simulation_engine.announcement import Announcement as Ann

if TYPE_CHECKING:
    from bgpsimulator.shared import Prefix


class RoSTTrustedRepository:
    def __init__(self) -> None:
        # {prefix: {id: active?}}
        # Since we don't do batches in sims, id is just asn
        self.info: defaultdict[Prefix, dict[int, bool]] = defaultdict(dict)

    def __repr__(self) -> str:
        return str(self.info)

    def clear(self) -> None:
        self.__init__()  # type: ignore

    def add_ann(self, ann: Ann, asn: int, active: bool) -> None:
        self.info[ann.prefix][asn] = active

    def seen_withdrawal(self, adj_ribs_in_ann: Ann) -> bool:
        rost_id_statuses_for_prefix = self.info[adj_ribs_in_ann.prefix]
        for rost_id in adj_ribs_in_ann.rost_ids:
            # This checks if it's active or not. Active=True
            if rost_id_statuses_for_prefix[rost_id] is False:
                return True
        return False



from typing import TYPE_CHECKING

from bgpsimulator.shared.enums import Relationships, Settings

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy

from .rov import ROV


class PathEnd:
    """A Policy that deploys Path-End

    Jump starting BGP with Path-End validation
    """

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Path-End extends ROV by checking the next-hop of the origin"""

        if not ROV.valid_ann(policy, ann, from_rel):
            return False

        origin_asn = ann.origin
        origin_as_obj = policy.as_.as_graph.as_dict.get(origin_asn)
        # If the origin is deploying pathend and the path is longer than 1
        if (
            origin_as_obj
            and origin_as_obj.policy.settings[Settings.PATH_END]
            and len(ann.as_path) > 1
        ):
            # If the provider is real, do the loop check
            for neighbor_asn in origin_as_obj.neighbor_asns:
                if neighbor_asn == ann.as_path[-2]:
                    return True
            # Provider is fake, return False
            return False
        else:
            return True



import math
import random
from functools import cached_property
from typing import TYPE_CHECKING, ClassVar

from bgpsimulator.route_validator import ROA, RouteValidator
from bgpsimulator.shared import IPAddr, Settings
from bgpsimulator.simulation_engine import Announcement as Ann
from bgpsimulator.simulation_engine import SimulationEngine

if TYPE_CHECKING:
    from bgpsimulator.as_graphs import AS

    from .scenario_config import ScenarioConfig


class Scenario:
    """Contains information regarding a scenario/attack

    This represents a single trial and a single engine run
    """

    min_propagation_rounds: int = 1

    # Used when dumping the scenario_config to JSON
    name_to_cls_dict: ClassVar[dict[str, type["Scenario"]]] = {}

    def __init_subclass__(cls, **kwargs):
        """Used when dumping the scenario_config to JSON

        NOTE: When converting to rust, just keep a hardcoded list of scenarios
        """
        super().__init_subclass__(**kwargs)
        Scenario.name_to_cls_dict[cls.__name__] = cls

    def __init__(
        self,
        *,
        scenario_config: "ScenarioConfig",
        engine: SimulationEngine,
        route_validator: RouteValidator,
        percent_ases_randomly_adopting: float = 0,
        attacker_asns: set[int] | None = None,
        legitimate_origin_asns: set[int] | None = None,
        adopting_asns: set[int] | None = None,
    ):
        """inits attrs

        Any kwarg prefixed with default is only required for the test suite/YAML
        """

        # Config's ScenarioCls must be the same as instantiated Scenario
        assert scenario_config.ScenarioCls == self.__class__, (
            "The config's scenario class is "
            f"{scenario_config.ScenarioCls.__name__}, but the scenario used is "
            f"{self.__class__.__name__}"
        )

        self.scenario_config: ScenarioConfig = scenario_config
        self.percent_ases_randomly_adopting: float = percent_ases_randomly_adopting

        self.attacker_asns: set[int] = self._get_attacker_asns(
            scenario_config.override_attacker_asns,
            attacker_asns,
            engine,
        )

        self.legitimate_origin_asns: set[int] = self._get_legitimate_origin_asns(
            scenario_config.override_legitimate_origin_asns,
            legitimate_origin_asns,
            engine,
        )
        self.adopting_asns: set[int] = self._get_adopting_asns(
            scenario_config.override_adopting_asns,
            adopting_asns,
            engine,
        )

        if self.scenario_config.override_seed_asn_ann_dict is not None:
            self.seed_asn_ann_dict: dict[int, list[Ann]] = (
                self.scenario_config.override_seed_asn_ann_dict.copy()
            )
        else:
            self.seed_asn_ann_dict = self._get_seed_asn_ann_dict(engine=engine)

        if self.scenario_config.override_roas is not None:
            self.roas: list[ROA] = self.scenario_config.override_roas.copy()
        else:
            self.roas = self._get_roas(
                seed_asn_ann_dict=self.seed_asn_ann_dict, engine=engine
            )
        self._reset_and_add_roas_to_route_validator(route_validator)

        if self.scenario_config.override_dest_ip_addr is not None:
            self.dest_ip_addr: IPAddr = self.scenario_config.override_dest_ip_addr
        else:
            self.dest_ip_addr = self._get_dest_ip_addr()

    def _reset_and_add_roas_to_route_validator(self, route_validator) -> None:
        """Clears & adds ROAs to route_validator.

        Which serves as RPKI+Routinator combo.
        """

        route_validator.clear()
        for roa in self.roas:
            route_validator.add_roa(roa)

    #################
    # Get attackers #
    #################

    def _get_attacker_asns(
        self,
        override_attacker_asns: set[int] | None,
        prev_attacker_asns: set[int] | None,
        engine: SimulationEngine,
    ) -> set[int]:
        """Returns attacker ASN at random"""

        # This is hardcoded, do not recalculate
        if override_attacker_asns is not None:
            attacker_asns = override_attacker_asns.copy()
            branch = 0
        # Reuse the attacker from the last scenario for comparability
        elif (
            prev_attacker_asns
            and len(prev_attacker_asns) == self.scenario_config.num_attackers
        ):
            attacker_asns = prev_attacker_asns
            branch = 1
        # This is being initialized for the first time, or old scenario has a
        # different number of attackers
        else:
            branch = 3
            assert engine
            possible_attacker_asns = self._get_possible_attacker_asns(
                engine, self.percent_ases_randomly_adopting
            )

            assert len(possible_attacker_asns) >= self.scenario_config.num_attackers
            # https://stackoverflow.com/a/15837796/8903959
            attacker_asns = set(
                random.sample(
                    tuple(possible_attacker_asns), self.scenario_config.num_attackers
                )
            )

        # Validate attacker asns
        err = f"Number of attackers is different from attacker length: Branch {branch}"
        assert len(attacker_asns) == self.scenario_config.num_attackers, err

        return attacker_asns

    def _get_possible_attacker_asns(
        self,
        engine: SimulationEngine,
        percent_ases_randomly_adopting: float,
    ) -> set[int]:
        """Returns possible attacker ASNs, defaulted from config"""

        possible_asns = engine.as_graph.asn_groups[
            self.scenario_config.attacker_asn_group
        ]
        return possible_asns

    ###############
    # Get Victims #
    ###############

    def _get_legitimate_origin_asns(
        self,
        override_legitimate_origin_asns: set[int] | None,
        prev_legitimate_origin_asns: set[int] | None,
        engine: SimulationEngine,
    ) -> set[int]:
        """Returns legitimate_origin ASN at random"""

        # This is coming from YAML, do not recalculate
        if override_legitimate_origin_asns is not None:
            legitimate_origin_asns = override_legitimate_origin_asns.copy()
        # Reuse the legitimate_origin from the last scenario for comparability
        elif (
            prev_legitimate_origin_asns
            and len(prev_legitimate_origin_asns)
            == self.scenario_config.num_legitimate_origins
        ):
            legitimate_origin_asns = prev_legitimate_origin_asns
        # This is being initialized for the first time
        else:
            assert engine
            possible_legitimate_origin_asns = self._get_possible_legitimate_origin_asns(
                engine, self.percent_ases_randomly_adopting
            )
            # https://stackoverflow.com/a/15837796/8903959
            legitimate_origin_asns = set(
                random.sample(
                    tuple(possible_legitimate_origin_asns),
                    self.scenario_config.num_legitimate_origins,
                )
            )

        err = (
            "Number of legitimate_origins is different from "
            f"legitimate_origin length {len(legitimate_origin_asns)} "
            f"{self.scenario_config.num_legitimate_origins}"
        )
        assert (
            len(legitimate_origin_asns) == self.scenario_config.num_legitimate_origins
        ), err

        return legitimate_origin_asns

    def _get_possible_legitimate_origin_asns(
        self,
        engine: SimulationEngine,
        percent_ases_randomly_adopting: float,
    ) -> set[int]:
        """Returns possible legitimate_origin ASNs, defaulted from config"""

        possible_asns = engine.as_graph.asn_groups[
            self.scenario_config.legitimate_origin_asn_group
        ]
        # Remove attackers from possible legitimate_origins
        possible_asns = possible_asns.difference(self.attacker_asns)
        return possible_asns

    #######################
    # Adopting ASNs funcs #
    #######################

    def _get_adopting_asns(
        self,
        override_adopting_asns: set[int] | None,
        adopting_asns: set[int] | None,
        engine: SimulationEngine,
    ) -> set[int]:
        """Returns all asns that will be adopting self.AdoptPolicyCls"""

        if override_adopting_asns is not None:
            return override_adopting_asns.copy()
        # By default use the same adopting ASes as the last scenario config
        elif adopting_asns:
            return adopting_asns
        else:
            adopting_asns = self._get_randomized_adopting_asns(engine)

        return adopting_asns

    def _get_randomized_adopting_asns(
        self,
        engine: SimulationEngine,
    ) -> set[int]:
        """Returns the set of adopting ASNs (aside from hardcoded ASNs)"""

        adopting_asns: list[int] = list()
        # Randomly adopt in all three subcategories
        for subcategory in self.scenario_config.adoption_asn_groups:
            asns = engine.as_graph.asn_groups[subcategory]
            # Remove ASes that are already pre-set
            # Ex: Attackers and legitimate_origins,
            # self.scenario_config.hardcoded_asn_cls_dict
            possible_adopters = asns.difference(self.preset_asns)

            # Get how many ASes should be adopting (store as k)
            if self.percent_ases_randomly_adopting == 0:
                k = 0
            else:
                k = math.ceil(
                    len(possible_adopters) * self.percent_ases_randomly_adopting / 100
                )

            try:
                # https://stackoverflow.com/a/15837796/8903959
                adopting_asns.extend(random.sample(tuple(possible_adopters), k))
            except ValueError as e:
                raise ValueError(
                    f"{k} can't be sampled from {len(possible_adopters)}"
                ) from e
        return set(adopting_asns)

    @cached_property
    def default_adopters(self) -> set[int]:
        """By default, legitimate_origin always adopts"""

        return self.legitimate_origin_asns

    @cached_property
    def default_non_adopters(self) -> set[int]:
        """By default, attacker always does not adopt"""

        return self.attacker_asns

    @cached_property
    def preset_asns(self) -> set[int]:
        """ASNs that have a preset adoption policy"""

        # Returns the union of default adopters and non adopters
        hardcoded_asns = set(self.scenario_config.override_adoption_settings)
        return self.default_adopters | self.default_non_adopters | hardcoded_asns

    @property
    def untracked_asns(self) -> set[int]:
        """Returns ASNs that shouldn't be tracked by the graphing tools

        By default just the default adopters and non adopters, and hardcoded ASNs
        """

        return self.default_adopters | self.default_non_adopters

    #############################
    # Engine Manipulation Funcs #
    #############################

    def set_settings(self, as_obj: "AS") -> None:
        """Sets the routing policy settings for a given AS"""

        # NOTE: Most important updates go last

        settings = [False for _ in Settings]

        for setting, val in self.scenario_config.default_base_settings.items():
            settings[setting] = val

        for setting, val in self.scenario_config.override_base_settings.get(
            as_obj.asn, {}
        ).items():
            settings[setting] = val

        asn = as_obj.asn
        if asn in self.scenario_config.override_adoption_settings:
            for setting, val in self.scenario_config.override_adoption_settings[
                asn
            ].items():
                settings[setting] = val
        elif asn in self.adopting_asns or asn in self.default_adopters:
            for setting, val in self.scenario_config.default_adoption_settings.items():
                settings[setting] = val

        if asn in self.attacker_asns:
            for setting, val in self.scenario_config.attacker_settings.items():
                settings[setting] = val
        elif asn in self.legitimate_origin_asns:
            for setting, val in self.scenario_config.legitimate_origin_settings.items():
                settings[setting] = val

        as_obj.policy.settings = tuple(settings)

    def setup_engine(self, engine: SimulationEngine) -> None:
        """Nice hook func for setting up the engine.

        With adopting ASes, routing policy settings, etc.
        """
        engine.setup(self)

    ##################
    # Subclass Funcs #
    ##################

    def _get_seed_asn_ann_dict(self, engine: SimulationEngine) -> dict[int, list[Ann]]:
        """Returns a dict of ASNs to announcements

        Empty by default for testing, typically subclassed
        """

        return {}

    def _get_roas(
        self,
        *,
        seed_asn_ann_dict: dict[int, list[Ann]],
        engine: SimulationEngine,
    ) -> list[ROA]:
        """Returns a list of ROA's

        Not abstract and by default does nothing for
        backwards compatability
        """
        return []

    def _get_dest_ip_addr(self) -> IPAddr:
        """Returns the destination IP address for the scenario

        Subclass must implement this
        """

        raise NotImplementedError("Subclass must implement this")

    def pre_aggregation_hook(
        self,
        engine: SimulationEngine,
        percent_ases_randomly_adopting: float,
        trial: int,
        propagation_round: int,
    ) -> None:
        """Useful hook for changes/checks
        prior to results aggregation.
        """
        pass

    def post_propagation_hook(
        self,
        engine: SimulationEngine,
        percent_ases_randomly_adopting: float,
        trial: int,
        propagation_round: int,
    ) -> None:
        """Useful hook for post propagation"""

        pass

    # NOTE: No JSON funcs since you can't store the engine


Scenario.name_to_cls_dict["Scenario"] = Scenario



from enum import Enum, IntEnum, StrEnum

from .prefix import Prefix


class Relationships(IntEnum):
    """Relationships between ASes.

    Higher numbers == higher priority for announcements
    """

    # Must start at one for the priority
    PROVIDERS = 1
    PEERS = 2
    # Customers have highest priority
    # Economic incentives first!
    CUSTOMERS = 3
    # Origin must always remain since the AS created it
    ORIGIN = 4
    # Unknown for external programs like extrapolator
    UNKNOWN = 5


class ASNGroups(StrEnum):
    """ASN groups"""

    TIER_1 = "tier_1"
    ETC = "etc"
    STUBS_OR_MH = "stubs_or_mh"
    ALL_WOUT_IXPS = "all_wout_ixps"
    STUBS = "stubs"
    MULTIHOMED = "multihomed"
    TRANSIT = "transit"
    IXPS = "ixps"


class Settings(IntEnum):
    """Routing policy settings"""

    ANNOUNCE_THEN_WITHDRAW = 0
    AS_PATH_EDGE_FILTER = 1
    ASPA = 2
    ASPA_W_N = 3
    ASPAPP = 4
    ASRA = 5
    BGPSEC = 6
    BGP_FULL = 7
    BGP_I_SEC = 8
    BGP_I_SEC_TRANSITIVE = 9
    ENFORCE_FIRST_AS = 10
    FIRST_ASN_STRIPPING_PREFIX_HIJACK_CUSTOMERS = 11
    LEAKER = 12
    NEVER_PROPAGATE_WITHDRAWALS = 13
    NEVER_WITHDRAW = 14
    ONLY_TO_CUSTOMERS = 15
    ORIGIN_PREFIX_HIJACK_CUSTOMERS = 16
    PATH_END = 17
    PEERLOCK_LITE = 18
    PEER_ROV = 19
    PROVIDER_CONE_ID = 20
    ROST = 21
    ROV = 22
    ROVPP_V1_LITE = 23
    ROVPP_V2I_LITE = 24
    ROVPP_V2_LITE = 25


class ROAValidity(IntEnum):
    """ROAValidity values

    NOTE: it's possible that you could have two ROAs for
    the same prefix, each with different reasons why they are
    invalid. In that case, the ROAChecker returns the "best"
    validity (in the order below). It doesn't really matter,
    since they are both invalid anyways, and that's the only
    case where this conflict can occur
    """

    # NOTE: These values double as "scores" for validity,
    # so do NOT change the orders
    # (used in the ROA class)
    VALID = 0
    UNKNOWN = 1
    INVALID_LENGTH = 2
    INVALID_ORIGIN = 3
    INVALID_LENGTH_AND_ORIGIN = 4

    @staticmethod
    def is_valid(roa_validity: "ROAValidity") -> bool:
        return roa_validity == ROAValidity.VALID

    @staticmethod
    def is_unknown(roa_validity: "ROAValidity") -> bool:
        return roa_validity == ROAValidity.UNKNOWN

    @staticmethod
    def is_invalid(roa_validity: "ROAValidity") -> bool:
        return roa_validity in (
            ROAValidity.INVALID_LENGTH,
            ROAValidity.INVALID_ORIGIN,
            ROAValidity.INVALID_LENGTH_AND_ORIGIN,
        )


class ROARouted(IntEnum):
    ROUTED = 0
    UNKNOWN = 1
    # A ROA is Non Routed if it is for an origin of ASN 0
    # This means that the prefix for this ROA should never be announced
    NON_ROUTED = 2


class CommonPrefixes(Enum):
    """Prefixes for attacks.

    Prefixes tend to be in reference to the legitimate origin.
    """

    SUPERPREFIX = Prefix("1.0.0.0/8")
    PREFIX = Prefix("1.2.0.0/16")
    SUBPREFIX = Prefix("1.2.3.0/24")


class Timestamps(IntEnum):
    """Timestamps for announcements. Legitimate origin is always first."""

    LEGITIMATE_ORIGIN = 0
    ATTACKER = 1


class Outcomes(IntEnum):
    ATTACKER_SUCCESS = 0
    LEGITIMATE_ORIGIN_SUCCESS = 1
    DISCONNECTED = 2
    UNDETERMINED = 3
    DATA_PLANE_LOOP = 4


class InAdoptingASNs(StrEnum):
    TRUE = "True"
    FALSE = "False"
    ANY = "Any"


class CommonASNs(IntEnum):
    """Common ASNs"""

    ATTACKER = 666
    LEGITIMATE_ORIGIN = 777



from pathlib import Path

import pytest

from bgpsimulator.engine_runner import EngineRunConfig, EngineRunner

from .engine_test_configs import engine_test_configs


@pytest.mark.engine
class TestEngine:
    """Performs a system test on the engine

    See tutorial for in depth details
    """

    @pytest.mark.parametrize("conf", engine_test_configs)
    def test_engine(self, conf: EngineRunConfig, overwrite: bool, dpi: int | None):
        """Performs a system test on the engine

        See tutorial for in depth details
        """

        EngineRunner(
            base_dir=self.base_dir,
            engine_run_config=conf,
            overwrite=overwrite,
            compare_against_ground_truth=True,
            write_diagrams=True,
        ).run(dpi=dpi)

    @property
    def base_dir(self) -> Path:
        """Returns test output dir"""

        return Path(__file__).parent / "engine_test_outputs"



from typing import TYPE_CHECKING

from bgpsimulator.shared import PolicyPropagateInfo, Relationships

if TYPE_CHECKING:
    from bgpsimulator.as_graphs import AS
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class OriginPrefixHijackCustomers:
    """A Policy that performs prefix origin hijacks against customers

    This is particularly useful against ASPA, which doesn't validate anns
    coming from providers. ASRA and ASPAwN thwards this, as do some other policies
    """

    @staticmethod
    def get_policy_propagate_vals(
        policy: "Policy",
        neighbor_as: "AS",
        ann: "Ann",
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> PolicyPropagateInfo:
        """Performs origin hijacks against customers"""

        # This ann is originating from here, the attacker, so it's an attacker's ann
        # If as path length is 1 (like it would be against BGP), don't modify it
        if (
            propagate_to == Relationships.CUSTOMERS
            and ann.recv_relationship == Relationships.ORIGIN
            and len(ann.as_path) > 1
        ):
            # Only need origin hijack when sending to customers
            return PolicyPropagateInfo(
                policy_propagate_bool=True,
                ann=ann.copy(as_path=(policy.as_.asn, ann.origin)),
                send_ann_bool=True,
            )
        else:
            return PolicyPropagateInfo(
                policy_propagate_bool=False,
                ann=ann,
                send_ann_bool=True,
            )



import json
from pathlib import Path
from typing import Any, Callable

from frozendict import frozendict

from bgpsimulator.as_graphs.as_graph import AS, ASGraphUtils
from bgpsimulator.shared import SINGLE_DAY_CACHE_DIR, bgpsimulator_logger
from bgpsimulator.simulation_engine.policy import Policy

from .caida_as_graph_collector import CAIDAASGraphCollector


class CAIDAASGraphJSONConverter:
    """Converts the serial-2 to a JSON file that can be ingested to create a graph"""

    def __init__(self, cache_dir: Path = SINGLE_DAY_CACHE_DIR) -> None:
        self.cache_dir: Path = cache_dir

    def run(
        self,
        caida_as_graph_path: Path | None = None,
        additional_asn_group_filters: frozendict[
            str, Callable[[dict[int, AS]], frozenset[int]]
        ] = frozendict(),
        PolicyCls: type[Policy] = Policy,
    ) -> tuple[dict[str, Any], Path]:
        """Generates AS graph in the following steps:

        1. download file from source using the GraphCollector if you haven't already
        2. Convert the CAIDA file to JSON
        3. Write to JSON path if it is set
        4. Return JSON info for speed (rather than rereading it later)
        """

        caida_as_graph_path = caida_as_graph_path or CAIDAASGraphCollector().run()
        # /home/anon/.cache/bgpsimulator/2025-06-24/
        # CAIDAASGraphCollector_2025.06.14.json
        json_cache_path = caida_as_graph_path.with_suffix(".json")
        if not json_cache_path.exists():
            self._write_as_graph_json(
                caida_as_graph_path,
                json_cache_path,
                additional_asn_group_filters,
                PolicyCls,
            )
        try:
            as_graph_info = json.loads(json_cache_path.read_text())
            # Must convert keys to ints when coming from JSON
            as_graph_info["ases"] = {
                int(asn): info for asn, info in as_graph_info["ases"].items()
            }
            return as_graph_info, json_cache_path
        except json.JSONDecodeError:
            bgpsimulator_logger.error(
                f"JSON file {json_cache_path} is corrupted, it will now be deleted"
            )
            json_cache_path.unlink()
            raise

    def _write_as_graph_json(
        self,
        caida_as_graph_path: Path,
        json_cache_path: Path,
        additional_asn_group_filters: frozendict[
            str, Callable[[dict[int, AS]], frozenset[int]]
        ],
        PolicyCls: type[Policy],
    ) -> None:
        """Writes as graph JSON from CAIDAs raw file"""

        asn_to_as: dict[int, AS] = dict()
        lines = caida_as_graph_path.read_text().splitlines()
        for line in lines:
            # Get Caida input clique. See paper on site for what this is
            if line.startswith("# input clique"):
                self._extract_tier_1_asns(line, asn_to_as)
            # Get detected Caida IXPs. See paper on site for what this is
            elif line.startswith("# IXP ASes"):
                self._extract_ixp_asns(line, asn_to_as)
            # Not a comment, must be a relationship
            elif not line.startswith("#"):
                # Extract all customer provider pairs
                if "-1" in line:
                    self._extract_provider_customers(line, asn_to_as)
                # Extract all peers
                else:
                    self._extract_peers(line, asn_to_as)

        final_json: dict[str, Any] = {
            "ases": {k: as_.to_json() for k, as_ in asn_to_as.items()},
        }
        ASGraphUtils.add_extra_setup(final_json, additional_asn_group_filters)

        with json_cache_path.open("w") as f:
            # add separators to make JSON as short as possible
            # ensure_ascii set to false also gives a speed boost
            json.dump(final_json, f, separators=(",", ":"), ensure_ascii=False)

    #################
    # Parsing funcs #
    #################

    def _extract_tier_1_asns(self, line: str, asn_to_as: dict[int, AS]) -> None:
        """Adds all ASNs within input clique line to ases dict"""

        # Gets all input ASes for clique
        for asn in line.split(":")[-1].strip().split(" "):
            as_ = asn_to_as.setdefault(int(asn), AS(asn=int(asn)))
            as_.tier_1 = True

    def _extract_ixp_asns(self, line: str, asn_to_as: dict[int, AS]) -> None:
        """Adds all ASNs that are detected IXPs to ASes dict"""

        # Get all IXPs that Caida lists
        for asn in line.split(":")[-1].strip().split(" "):
            as_ = asn_to_as.setdefault(int(asn), AS(asn=int(asn)))
            as_.ixp = True

    def _extract_provider_customers(self, line: str, asn_to_as: dict[int, AS]) -> None:
        """Extracts provider customers: <provider-as>|<customer-as>|-1"""

        provider_asn, customer_asn, _, source = line.split("|")

        provider_as = asn_to_as.setdefault(int(provider_asn), AS(asn=int(provider_asn)))
        provider_as.customer_asns.add(int(customer_asn))

        customer_as = asn_to_as.setdefault(int(customer_asn), AS(asn=int(customer_asn)))
        customer_as.provider_asns.add(int(provider_asn))

    def _extract_peers(self, line: str, asn_to_as: dict[int, AS]) -> None:
        """Extracts peers: <peer-as>|<peer-as>|0|<source>"""

        peer1_asn, peer2_asn, _, source = line.split("|")

        peer1_as = asn_to_as.setdefault(int(peer1_asn), AS(asn=int(peer1_asn)))
        peer1_as.peer_asns.add(int(peer2_asn))

        peer2_as = asn_to_as.setdefault(int(peer2_asn), AS(asn=int(peer2_asn)))
        peer2_as.peer_asns.add(int(peer1_asn))



from typing import TYPE_CHECKING

from bgpsimulator.shared import Relationships

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class ROST:
    """A Policy that deploys ROST"""

    @staticmethod
    def withdraw_ann_from_neighbors(policy: "Policy", withdraw_ann: "Ann") -> None:
        """Adds withdrawals you create to RoST Trusted Repo"""

        policy.rost_trusted_repository.add_ann(
            withdraw_ann, policy.as_.asn, active=False
        )

    @staticmethod
    def preprocess_incoming_anns(
        policy: "Policy",
        *,
        from_rel: Relationships,
        propagation_round: int = 0,
        **kwargs,
    ) -> None:
        """Adds withdrawals and anns you received to the rost trusted repo"""

        ROST.remove_anns_from_recv_q_that_should_be_withdrawn(policy)
        ROST.add_suppressed_withdrawals_back_to_recv_q(policy)

    @staticmethod
    def postprocess_incoming_anns(policy: "Policy") -> None:
        """sets local rib anns to active in rost trusted repo"""

        for ann in policy.local_rib.values():
            policy.rost_trusted_repository.add_ann(ann, policy.as_.asn, active=True)

    @staticmethod
    def remove_anns_from_recv_q_that_should_be_withdrawn(policy: "Policy") -> None:
        for prefix, ann_list in policy.recv_q.copy().items():
            new_ann_list = list()
            for ann in ann_list:
                # So long as it's not a new ann that has a suppressed withdrawal
                # add to recv_q
                if not (
                    ann.withdraw is False
                    and policy.rost_trusted_repository.seen_withdrawal(ann)
                ):
                    new_ann_list.append(ann)
            policy.recv_q[prefix] = new_ann_list

    @staticmethod
    def add_suppressed_withdrawals_back_to_recv_q(policy: "Policy") -> None:
        for inner_dict in policy.adj_ribs_in.values():
            for ann_info in inner_dict.values():
                adj_ribs_in_ann = ann_info.unprocessed_ann

                # Determine if the AdjRIBsIn ann is already being withdrawn
                withdrawal_in_recv_q = False
                for ann in policy.recv_q.get(adj_ribs_in_ann.prefix, []):
                    if adj_ribs_in_ann.as_path == ann.as_path:
                        withdrawal_in_recv_q = True

                # if adj_ribs_in_ann withdrawal not in the recv_q,
                # and adj_ribs_in_ann in the trusted repo, create a withdrawal
                if (
                    not withdrawal_in_recv_q
                    and policy.rost_trusted_repository.seen_withdrawal(adj_ribs_in_ann)
                ):
                    policy.recv_q[adj_ribs_in_ann.prefix].append(
                        adj_ribs_in_ann.copy(withdraw=True)
                    )

    @staticmethod
    def process_ann(policy: "Policy", ann: "Ann", from_rel: "Relationships") -> "Ann":
        """Processes an announcement for RoST"""

        return ann.copy(rost_ids=(policy.as_.asn, *ann.rost_ids))



from typing import TYPE_CHECKING

from bgpsimulator.shared import Relationships

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine.policy.policy import Policy
    from bgpsimulator.simulation_framework import Scenario


class Leaker:
    """Leaks everything second round of propagation

    Generally this is more efficient to do with the ScenarioCls
    of AccidentalRouteLeak but this is helpful for the website
    """

    @staticmethod
    def pre_propagation_hook(
        policy: "Policy", propagation_round: int, scenario: "Scenario"
    ) -> None:
        """Leak everything second round of propagation (which is zero indexed)"""

        if propagation_round == 1:
            for ann in list(policy.local_rib.values()):
                if ann.recv_relationship != Relationships.ORIGIN:
                    new_ann = policy.local_rib.pop(ann.prefix).copy(
                        recv_relationship=Relationships.CUSTOMERS
                    )
                    policy.local_rib[new_ann.prefix] = new_ann



from bgpsimulator.as_graphs.as_graph import ASGraph
from bgpsimulator.engine_runner import EngineRunConfig
from bgpsimulator.shared import CommonASNs
from bgpsimulator.simulation_framework import ScenarioConfig, SubprefixHijack

graph_data = {
    "ases": {
        str(CommonASNs.LEGITIMATE_ORIGIN): {
            "asn": CommonASNs.LEGITIMATE_ORIGIN,
            "customer_asns": [],
            "peer_asns": [],
            "provider_asns": [2, 4, 10],
        },
        str(CommonASNs.ATTACKER): {
            "asn": CommonASNs.ATTACKER,
            "customer_asns": [],
            "peer_asns": [],
            "provider_asns": [1, 2],
        },
        "1": {
            "asn": 1,
            "customer_asns": [CommonASNs.ATTACKER],
            "peer_asns": [],
            "provider_asns": [5, 8],
        },
        "2": {
            "asn": 2,
            "customer_asns": [CommonASNs.ATTACKER, CommonASNs.LEGITIMATE_ORIGIN],
            "peer_asns": [],
            "provider_asns": [8],
        },
        "3": {
            "asn": 3,
            "customer_asns": [],
            "peer_asns": [9],
            "provider_asns": [],
        },
        "4": {
            "asn": 4,
            "customer_asns": [CommonASNs.LEGITIMATE_ORIGIN],
            "peer_asns": [],
            "provider_asns": [9],
        },
        "5": {
            "asn": 5,
            "customer_asns": [1],
            "peer_asns": [],
            "provider_asns": [],
        },
        "8": {
            "asn": 8,
            "customer_asns": [1, 2],
            "peer_asns": [9],
            "provider_asns": [11],
        },
        "9": {
            "asn": 9,
            "customer_asns": [4],
            "peer_asns": [8, 10, 3],
            "provider_asns": [11],
        },
        "10": {
            "asn": 10,
            "customer_asns": [CommonASNs.LEGITIMATE_ORIGIN],
            "peer_asns": [9],
            "provider_asns": [11, 12],
        },
        "11": {
            "asn": 11,
            "customer_asns": [8, 9, 10],
            "peer_asns": [],
            "provider_asns": [],
        },
        "12": {
            "asn": 12,
            "customer_asns": [10],
            "peer_asns": [],
            "provider_asns": [],
        },
    },
}

desc = (
    "Subprefix hijack with BGP Simple "
    "Valley Free (Gao Rexford) Demonstration\n"
    "import policy\n"
    "AS 9, prefix, shows customer > peer\n"
    "AS 9, subprefix, shows peer > provider\n"
    "AS 11, prefix, shows shortest AS path\n"
    "AS 5 and AS 8, subprefix, tiebreaker by lowest ASN\n"
    "export policy\n"
    "AS 10, subprefix, shows anns from providers only export to customers\n"
    "AS 9, subprefix, shows anns from peers only export to customers\n"
    "(All ASes show exporting to customers)\n"
    "hidden hijack\n"
    "AS 12 shows a hidden hijack\n"
)

# Create the engine run config
ex_config_002 = EngineRunConfig(
    name="ex_002_subprefix_hijack_bgp_simple_gao_rexford_demo",
    scenario_config=ScenarioConfig(
        label="bgp",
        ScenarioCls=SubprefixHijack,
        override_attacker_asns={CommonASNs.ATTACKER},
        override_legitimate_origin_asns={CommonASNs.LEGITIMATE_ORIGIN},
    ),
    as_graph=ASGraph(graph_data),
    diagram_desc=desc,
)



from functools import cached_property
from typing import Any

from bgpsimulator.shared import Prefix, ROARouted, ROAValidity


class ROA:
    def __init__(
        self,
        prefix: Prefix,
        origin: int,
        max_length: int | None = None,
        ta: str | None = None,
    ):
        self.prefix: Prefix = prefix
        self.origin: int = origin
        self.max_length: int = max_length or self.prefix.prefixlen
        self.ta: str | None = ta

    def __hash__(self) -> int:
        return hash((self.prefix, self.origin, self.max_length, self.ta))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ROA):
            return (
                self.prefix == other.prefix
                and self.origin == other.origin
                and self.max_length == other.max_length
                and self.ta == other.ta
            )
        return NotImplemented

    @cached_property
    def routed_status(self) -> ROARouted:
        return ROARouted.ROUTED if self.is_routed else ROARouted.NON_ROUTED

    @cached_property
    def is_routed(self) -> bool:
        return self.origin != 0

    @cached_property
    def is_non_routed(self) -> bool:
        return not self.is_routed

    def covers_prefix(self, prefix: Prefix) -> bool:
        """Returns True if the ROA covers the prefix"""

        # NOTE: subnet_of includes the original prefix (I checked lol)
        return prefix.subnet_of(self.prefix)

    def get_validity(self, prefix: Prefix, origin: int) -> ROAValidity:
        """Returns validity of prefix origin pair"""

        if self.covers_prefix(prefix):
            if prefix.prefixlen > self.max_length and origin != self.origin:
                return ROAValidity.INVALID_LENGTH_AND_ORIGIN
            elif prefix.prefixlen > self.max_length and origin == self.origin:
                return ROAValidity.INVALID_LENGTH
            elif prefix.prefixlen <= self.max_length and origin != self.origin:
                return ROAValidity.INVALID_ORIGIN
            elif prefix.prefixlen <= self.max_length and origin == self.origin:
                return ROAValidity.VALID
            else:
                raise NotImplementedError("This should never happen")
        else:
            return ROAValidity.UNKNOWN

    def get_outcome(self, prefix: Prefix, origin: int) -> tuple[ROAValidity, ROARouted]:
        """Returns outcome of prefix origin pair"""

        return self.get_validity(prefix, origin), self.routed_status

    def to_json(self) -> dict[str, Any]:
        """Converts the ROA to a JSON object"""
        return {
            "prefix": str(self.prefix),
            "origin": self.origin,
            "max_length": self.max_length,
            "ta": self.ta,
        }

    @classmethod
    def from_json(cls, json_obj: dict[str, Any]) -> "ROA":
        """Converts a JSON object to a ROA"""
        return cls(
            prefix=Prefix(json_obj["prefix"]),
            origin=int(json_obj["origin"]),
            max_length=int(json_obj["max_length"]),
            ta=json_obj.get("ta"),
        )



from typing import TYPE_CHECKING

from bgpsimulator.shared.enums import Relationships

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class BGP:
    """A Policy that deploys BGP"""

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Determine if an announcement is valid or should be dropped"""

        # BGP Loop Prevention Check; no AS 0 either
        return policy.as_.asn not in ann.as_path and 0 not in ann.as_path



from typing import TYPE_CHECKING

from bgpsimulator.shared import (
    PolicyPropagateInfo,
    Relationships,
    ROARouted,
    ROAValidity,
)

from .rov import ROV
from .rovpp_v1_lite import ROVPPV1Lite

if TYPE_CHECKING:
    from bgpsimulator.as_graphs import AS
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class ROVPPV2Lite:
    """A Policy that deploys ROV++V2 Lite as defined in the ROV++ paper"""

    @staticmethod
    def process_incoming_anns(
        policy: "Policy", from_rel: Relationships, propagation_round: int
    ) -> None:
        """Additional processing for incoming announcements.

        The same as ROV++v1 (adding blackholes).
        """

        return ROVPPV1Lite.process_incoming_anns(policy, from_rel, propagation_round)

    @staticmethod
    def get_policy_propagate_vals(
        policy: "Policy",
        neighbor_as: "AS",
        ann: "Ann",
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> PolicyPropagateInfo:
        if ann.rovpp_blackhole:
            if ROVPPV2Lite.send_competing_hijack_allowed(policy, ann, propagate_to):
                return PolicyPropagateInfo(
                    policy_propagate_bool=True, ann=ann, send_ann_bool=True
                )
            else:
                return PolicyPropagateInfo(
                    policy_propagate_bool=True, ann=ann, send_ann_bool=False
                )
        else:
            return PolicyPropagateInfo(
                policy_propagate_bool=False, ann=ann, send_ann_bool=True
            )

    @staticmethod
    def send_competing_hijack_allowed(
        policy: "Policy", ann: "Ann", propagate_to: Relationships
    ) -> bool:
        """You can send blackhole to customers if from peer/provider.

        And either subprefix or non-routed.
        """

        roa_validity, roa_routed = policy.route_validator.get_roa_outcome(
            ann.prefix, ann.origin
        )
        # If we are the origin, then we can't send a competing hijack
        return (
            # From peer/provider
            ann.recv_relationship
            in (Relationships.PEERS, Relationships.PROVIDERS, Relationships.ORIGIN)
            # Sending to customers
            and propagate_to == Relationships.CUSTOMERS
            # subprefix or non routed (don't send blackholes for prefixes)
            # To tell if it's a subprefix hijack we check if it's invalid by length
            and (
                roa_validity
                not in (
                    ROAValidity.INVALID_LENGTH,
                    ROAValidity.INVALID_LENGTH_AND_ORIGIN,
                )
                or roa_routed == ROARouted.NON_ROUTED
            )
        )

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Determines ROV++V2 Lite validity"""

        return ROV.valid_ann(policy, ann, from_rel)



from .exceptions import InvalidIPAddressError
from .prefix import Prefix


class IPAddr(Prefix):
    """
    IPAddress class that wraps both IPv4 and IPv6 addresses,
    storing them internally as IPv6Network using IPv4-mapped format
    for IPv4 addresses (::ffff:a.b.c.d). Reserved addresses are disallowed.
    """

    __slots__ = ()

    def __init__(self, address: str):
        super().__init__(address)
        if self.prefixlen not in {32, 128}:
            raise InvalidIPAddressError(f"Invalid IP address: {address}")



from functools import cached_property
from typing import TYPE_CHECKING, Any, cast
from weakref import proxy

from bgpsimulator.shared import Relationships
from bgpsimulator.simulation_engine import Policy

if TYPE_CHECKING:
    from .as_graph import ASGraph


class AS:
    """Autonomous System class. Contains attributes of an AS"""

    def __init__(
        self,
        asn: int,
        peer_asns: set[int] | None = None,
        provider_asns: set[int] | None = None,
        customer_asns: set[int] | None = None,
        provider_cone_asns: set[int] | None = None,
        propagation_rank: int | None = None,
        tier_1: bool = False,
        ixp: bool = False,
        as_graph: "ASGraph | None" = None,
        policy_json: dict[str, Any] | None = None,
    ) -> None:
        # Make sure you're not accidentally passing in a string here
        self.asn: int = int(asn)

        self.peer_asns: set[int] = peer_asns or set()
        self.provider_asns: set[int] = provider_asns or set()
        self.customer_asns: set[int] = customer_asns or set()

        self.peers: list[AS] = []
        self.providers: list[AS] = []
        self.customers: list[AS] = []

        # Read Caida's paper to understand these
        self.tier_1: bool = tier_1
        self.ixp: bool = ixp
        self.provider_cone_asns: set[int] = provider_cone_asns or set()
        # Propagation rank. 0 is a leaf, highest is the input clique/t1 ASes
        self.propagation_rank: int | None = propagation_rank

        # Hash in advance and only once since this gets called a lot
        self.hashed_asn = hash(self.asn)

        self.policy: Policy = (
            Policy.from_json(policy_json, self) if policy_json else Policy(self)
        )

        # This is useful for some policies to have knowledge of the graph
        # Cast this here to avoid having to write the weakref everywhere
        if as_graph is not None:
            self.as_graph: ASGraph = cast("ASGraph", proxy(as_graph))
        else:
            # Ignoring this because it gets set properly immediatly
            self.as_graph = None  # type: ignore

    def set_relations(self) -> None:
        """Sets the relations for the AS"""
        if not self.as_graph:
            raise ValueError("AS graph not set")

        self.peers = [cast("AS", proxy(self.as_graph[asn])) for asn in self.peer_asns]
        self.providers = [
            cast("AS", proxy(self.as_graph[asn])) for asn in self.provider_asns
        ]
        self.customers = [
            cast("AS", proxy(self.as_graph[asn])) for asn in self.customer_asns
        ]

    def get_neighbor(self, rel: Relationships) -> list["AS"]:
        """Returns the neighbors of the AS based on the relationship enum"""

        if rel == Relationships.PEERS:
            return self.peers
        elif rel == Relationships.PROVIDERS:
            return self.providers
        elif rel == Relationships.CUSTOMERS:
            return self.customers
        else:
            raise ValueError(f"Invalid relationship: {rel}")

    def __lt__(self, as_obj: Any) -> bool:
        if isinstance(as_obj, AS):
            return self.asn < as_obj.asn
        else:
            return NotImplemented

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, AS):
            return self.to_json() == other.to_json()
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return self.hashed_asn

    @cached_property
    def stub(self) -> bool:
        """Returns True if AS is a stub by RFC1772

        Use neighbor_asns instead of neighbors so you can use this
        during graph construction
        """

        return len(self.neighbor_asns) == 1

    @cached_property
    def multihomed(self) -> bool:
        """Returns True if AS is multihomed by RFC1772

        Use customer_asns instead of customers so you can use this
        during graph construction
        """

        return (
            len(self.customer_asns) == 0
            and len(self.peer_asns) + len(self.provider_asns) > 1
        )

    @cached_property
    def transit(self) -> bool:
        """Returns True if AS is a transit AS by RFC1772

        Use customer_asns instead of customers so you can use this
        during graph construction
        """

        return (
            len(self.customer_asns) > 0
            and len(self.customer_asns) + len(self.peer_asns) + len(self.provider_asns)
            > 1
        )

    @cached_property
    def stubs(self) -> list["AS"]:
        """Returns a list of any stubs connected to that AS"""

        return [x for x in self.customers if x.stub]

    @cached_property
    def neighbors(self) -> list["AS"]:
        """Returns customers + peers + providers"""

        return self.customers + self.peers + self.providers

    @cached_property
    def neighbor_asns(self) -> set[int]:
        """Returns neighboring ASNs (useful for ASRA)"""

        return self.customer_asns | self.peer_asns | self.provider_asns

    ##############
    # JSON funcs #
    ##############

    def to_json(self) -> dict[str, Any]:
        """This optional method is called when you call yaml.dump()"""

        return {
            "asn": self.asn,
            "customer_asns": list(self.customer_asns),
            "peer_asns": list(self.peer_asns),
            "provider_asns": list(self.provider_asns),
            "tier_1": self.tier_1,
            "ixp": self.ixp,
            "provider_cone_asns": list(self.provider_cone_asns),
            "propagation_rank": self.propagation_rank,
            "policy": self.policy.to_json(),
        }

    @classmethod
    def from_json(
        cls, json_obj: dict[str, Any], as_graph: "ASGraph | None" = None
    ) -> "AS":
        """Converts the AS to a JSON object"""

        return cls(
            as_graph=as_graph,
            asn=int(json_obj["asn"]),
            customer_asns=set([int(x) for x in json_obj.get("customer_asns", [])]),
            peer_asns=set([int(x) for x in json_obj.get("peer_asns", [])]),
            provider_asns=set([int(x) for x in json_obj.get("provider_asns", [])]),
            tier_1=json_obj.get("tier_1", False),
            ixp=json_obj.get("ixp", False),
            provider_cone_asns=set(
                [int(x) for x in json_obj.get("provider_cone_asns", [])]
            ),
            propagation_rank=int(json_obj.get("propagation_rank"))
            if json_obj.get("propagation_rank")
            else None,
            policy_json=json_obj.get("policy", {}),
        )



import re
from pathlib import Path

from bgpsimulator.as_graphs import AS, ASGraph
from bgpsimulator.shared import ASNGroups, InAdoptingASNs, Outcomes
from bgpsimulator.simulation_framework.scenarios.scenario import Scenario


class LineFilter:
    """Filters for a line in a line chart

    Checks for each AS whether it meets the filter criteria, aside from
    the outcome

    The outcome is used only for the numerator of the data point, the
    denominator is always the total number of ASes in the group

    You can always subclass this and change your criteria for the numerator
    and denominator
    """

    __slots__ = ("asn_group", "in_adopting_asns", "prop_round", "outcome", "_hash")

    def __init__(
        self,
        as_group: ASNGroups,
        in_adopting_asns: InAdoptingASNs,
        prop_round: int,
        outcome: Outcomes,
    ) -> None:
        self.asn_group = as_group
        self.in_adopting_asns = in_adopting_asns
        self.prop_round = prop_round
        self.outcome = outcome
        self._hash = hash(self.to_json())

    def as_in_denominator(
        self,
        as_obj: AS,
        as_graph: ASGraph,
        scenario: Scenario,
        propagation_round: int,
        outcome: Outcomes,
    ) -> bool:
        """Checks if the AS meets the filter criteria, aside from the outcome

        The outcome is used only for the numerator of the data point, the
        denominator is always the total number of ASes in the group
        """
        if propagation_round != self.prop_round:
            return False
        elif as_obj.asn not in as_graph.asn_groups[self.asn_group]:
            return False
        elif (
            self.in_adopting_asns == InAdoptingASNs.TRUE
            and as_obj.asn not in scenario.adopting_asns
        ):
            return False
        elif (
            self.in_adopting_asns == InAdoptingASNs.FALSE
            and as_obj.asn in scenario.adopting_asns
        ):
            return False
        else:
            return True

    def as_in_numerator(
        self,
        as_obj: AS,
        as_graph: ASGraph,
        scenario: Scenario,
        propagation_round: int,
        outcome: Outcomes,
    ) -> bool:
        """Checks if the AS should be included in the numerator of the data point

        NOTE: as_in_denominator is already checked before this function is
        called, so we don't need to check it again
        """

        return outcome == self.outcome

    def __hash__(self) -> int:
        return self._hash

    def __repr__(self) -> str:
        return self.to_json()

    def __eq__(self, other) -> bool:
        if isinstance(other, LineFilter):
            return (
                self.asn_group == other.asn_group
                and self.in_adopting_asns == other.in_adopting_asns
                and self.prop_round == other.prop_round
                and self.outcome == other.outcome
            )
        else:
            return NotImplemented

    def to_json(self) -> str:
        """Returns a JSON-friendly string that can be used as a key"""
        return (
            f"All ASes in AS Group({self.asn_group}) "
            f"where adopting is set to ({self.in_adopting_asns}) "
            f"and propagation round is ({self.prop_round}) "
            f"and outcome is ({self.outcome})"
        )

    def to_csv(self) -> str:
        """Returns a CSV-friendly string"""
        return (
            f"{self.asn_group},{self.in_adopting_asns},"
            f"{self.prop_round},{self.outcome.name}"
        )

    @classmethod
    def from_json(cls, string: str) -> "LineFilter":
        """Converts a JSON-friendly string back to a LineFilter"""
        matches = re.findall(r"\((.*?)\)", string)
        if len(matches) != 4:
            raise ValueError(
                f"Expected 4 values in parentheses, got {len(matches)}: {matches}"
            )
        as_group, in_adopting_asns, prop_round, outcome = matches
        return cls(
            as_group=ASNGroups(as_group),
            in_adopting_asns=InAdoptingASNs(in_adopting_asns),
            prop_round=int(prop_round),
            outcome=Outcomes(int(outcome)),
        )

    def get_json_path(self, base_dir: Path) -> Path:
        path = (
            base_dir
            / "graph_jsons"
            / f"as_group_is_{self.asn_group.name}"
            / f"adopting_is_{self.in_adopting_asns.name}"
            / f"propagation_round_{self.prop_round}"
            / f"{self.outcome.name}.json"
        )
        path.parent.mkdir(exist_ok=True, parents=True)
        return path

    def get_png_path(self, base_dir: Path) -> Path:
        path = (
            base_dir
            / "graph_pngs"
            / f"as_group_is_{self.asn_group.name}"
            / f"adopting_is_{self.in_adopting_asns.name}"
            / f"propagation_round_{self.prop_round}"
            / f"{self.outcome.name}.png"
        )
        path.parent.mkdir(exist_ok=True, parents=True)
        return path



"""Test EngineRunConfig JSON serialization and deserialization"""

import json
from pathlib import Path

import pytest

from bgpsimulator.engine_runner import EngineRunConfig
from bgpsimulator.tests.engine_tests.engine_test_configs.examples.ex_config_000 import (
    ex_config_000,
)


class TestEngineRunConfig:
    """Tests for EngineRunConfig JSON roundtrip"""

    def test_json_roundtrip(self, tmp_path: Path):
        """Test that a config can be saved and loaded back identically"""

        config1 = ex_config_000

        # Convert to JSON and back
        json_str = json.dumps(config1.to_json(), indent=2)

        # Save to temporary file
        json_file_1 = tmp_path / "test_config_1.json"
        json_file_1.write_text(json_str)

        # Load it back
        json_obj = json.loads(json_str)
        json_obj["prevent_naming_duplicates"] = False  # Disable duplicate check
        config2 = EngineRunConfig.from_json(json_obj)

        # Convert config2 back to JSON
        json_str2 = json.dumps(config2.to_json(), indent=2)
        json_file_2 = tmp_path / "test_config_2.json"
        json_file_2.write_text(json_str2)

        # Parse both JSON strings for comparison
        json_obj_1 = json.loads(json_str)
        json_obj_2 = json.loads(json_str2)

        # The only difference should be prevent_naming_duplicates
        assert json_obj_1.get("prevent_naming_duplicates") == True
        assert json_obj_2.get("prevent_naming_duplicates") == False

        # Remove prevent_naming_duplicates for comparison
        json_obj_1.pop("prevent_naming_duplicates", None)
        json_obj_2.pop("prevent_naming_duplicates", None)

        # Now they should be identical
        assert json_obj_1 == json_obj_2

        # Config objects should be equal
        assert config1 == config2
        assert config1.scenario_config == config2.scenario_config
        assert config1.as_graph == config2.as_graph

    def test_propagation_rank_asns_preserved(self, tmp_path: Path):
        """Test that propagation_rank_asns is preserved during serialization"""

        config1 = ex_config_000

        # Convert to JSON
        json_obj = config1.to_json()

        # Check that propagation_rank_asns is present and not empty
        assert "as_graph" in json_obj
        assert "propagation_rank_asns" in json_obj["as_graph"]
        propagation_ranks = json_obj["as_graph"]["propagation_rank_asns"]
        assert len(propagation_ranks) > 0
        assert all(isinstance(rank, list) for rank in propagation_ranks)

        # Deserialize
        json_obj["prevent_naming_duplicates"] = False
        config2 = EngineRunConfig.from_json(json_obj)

        # Convert back to JSON
        json_obj2 = config2.to_json()

        # propagation_rank_asns should be preserved
        assert (
            json_obj["as_graph"]["propagation_rank_asns"]
            == json_obj2["as_graph"]["propagation_rank_asns"]
        )



from typing import TYPE_CHECKING

from bgpsimulator.shared import PolicyPropagateInfo, Relationships, Settings

if TYPE_CHECKING:
    from bgpsimulator.as_graphs import AS
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class BGPSec:
    """A Policy that deploys BGPSEC as defined in the BGPSEC RFC

    Since there are no real world implementations,
    we assume a secure path preference of security third,
    which is in line with the majority of users
    for the survey results in "A Survey of Interdomain Routing Policies"
    https://www.cs.bu.edu/~goldbe/papers/survey.pdf
    """

    @staticmethod
    def get_modified_seed_ann(policy: "Policy", ann: "Ann") -> "Ann":
        """Seeds an announcement into the local RIB and inits bgpsec_as_path"""

        # If the path is valid, add bgpsec_as_path
        if ann.as_path == (policy.as_.asn,):
            return ann.copy(bgpsec_as_path=ann.as_path)
        return ann

    @staticmethod
    def bgpsec_valid(policy: "Policy", ann: "Ann") -> bool:
        """Checks if an announcement is valid for BGPSEC"""
        return (
            ann.bgpsec_next_asn == policy.as_.asn and ann.bgpsec_as_path == ann.as_path
        )

    @staticmethod
    def get_policy_propagate_vals(
        policy: "Policy",
        neighbor_as: "AS",
        ann: "Ann",
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> PolicyPropagateInfo:
        """Gets the policy propagate values for a given announcement"""

        neighbor_bgpsec = any(
            neighbor_as.policy.settings[setting]
            for setting in [
                Settings.BGPSEC,
                Settings.BGP_I_SEC,
                Settings.BGP_I_SEC_TRANSITIVE,
            ]
        )
        if neighbor_bgpsec:
            next_asn = neighbor_as.asn
            path = ann.bgpsec_as_path
        else:
            next_asn = None
            path = ()
        send_ann = ann.copy(bgpsec_next_asn=next_asn, bgpsec_as_path=path)
        return PolicyPropagateInfo(
            policy_propagate_bool=True, ann=send_ann, send_ann_bool=True
        )

    @staticmethod
    def process_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> "Ann":
        """Sets the bgpsec_as_path. Prepends ASN if valid, otherwise clears"""

        # NOTE: must only look after the first ASN in the path, since the new ASN will
        # already have been prepended in the Policy class.
        if BGPSec.bgpsec_valid(policy, ann.copy(as_path=ann.as_path[1:])):
            return ann.copy(bgpsec_as_path=ann.as_path)
        else:
            return ann.copy(bgpsec_as_path=())

    @staticmethod
    def get_best_ann_by_bgpsec(
        policy: "Policy", current_ann: "Ann", new_ann: "Ann"
    ) -> "Ann | None":
        """Gets the best announcement by BGPSEC"""

        current_ann_valid = BGPSec.bgpsec_valid(policy, current_ann)
        new_ann_valid = BGPSec.bgpsec_valid(policy, new_ann)

        if current_ann_valid and not new_ann_valid:
            return current_ann
        elif not current_ann_valid and new_ann_valid:
            return new_ann
        else:
            return None



from typing import TYPE_CHECKING

from bgpsimulator.shared.enums import Relationships

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_engine.policy.policy import Policy


class ASPAPP:
    """A Policy that deploys ASPAPP"""

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Returns False if ann from peer/customer when ASPAPP is set"""

        raise NotImplementedError("ASPAPP is not supported yet")



from typing import TYPE_CHECKING, Any

from bgpsimulator.as_graphs import ASGraph
from bgpsimulator.shared import Relationships

if TYPE_CHECKING:
    from bgpsimulator.simulation_engine import Announcement as Ann
    from bgpsimulator.simulation_framework import Scenario


class SimulationEngine:
    """Python simulation engine representation"""

    def __init__(self, as_graph: "ASGraph") -> None:
        self.as_graph = as_graph

    def __eq__(self, other) -> bool:
        """Returns if two simulators contain the same BGPDAG's"""

        if isinstance(other, SimulationEngine):
            return self.as_graph == other.as_graph
        else:
            return NotImplemented

    ###############
    # Setup funcs #
    ###############

    def setup(self, scenario: "Scenario") -> None:
        """Sets AS classes and seeds announcements"""

        self._clear_as_routing_policies()
        self._set_settings(scenario)
        self._seed_announcements(scenario.seed_asn_ann_dict)

    def _clear_as_routing_policies(self) -> None:
        """Clears the routing policies for each AS"""

        for as_obj in self.as_graph:
            as_obj.policy.clear()

    def _set_settings(self, scenario: "Scenario") -> None:
        """Sets the routing policy settings for each AS"""

        for as_obj in self.as_graph:
            scenario.set_settings(as_obj)

    def _seed_announcements(self, seed_asn_ann_dict: dict[int, list["Ann"]]) -> None:
        """Seeds announcement at the proper AS

        Since this is the simulator engine, we should
        never have to worry about overlapping announcements
        """

        for asn, anns in seed_asn_ann_dict.items():
            for ann in anns:
                self.as_graph.as_dict[asn].policy.seed_ann(ann)

    #####################
    # Propagation funcs #
    #####################

    def propagate(self, propagation_round: int, scenario: "Scenario") -> None:
        """Propogates announcements

        to stick with Gao Rexford, we propagate to
        0. providers
        1. peers
        2. customers
        """

        self._pre_propagation_hook(propagation_round, scenario)
        self._propagate_to_providers(propagation_round, scenario)
        self._propagate_to_peers(propagation_round, scenario)
        self._propagate_to_customers(propagation_round, scenario)

    def _pre_propagation_hook(self, propagation_round, scenario) -> None:
        """Pre-propagation hook func in the policy

        Mainly for policies like LEAKER and ANNOUNCE_THEN_WITHDRAW
        """

        for as_obj in self.as_graph.as_dict.values():
            as_obj.policy.pre_propagation_hook(propagation_round, scenario)

    def _propagate_to_providers(
        self, propagation_round: int, scenario: "Scenario"
    ) -> None:
        """Propogate to providers"""

        # Propogation ranks go from stubs to input_clique in ascending order
        # By customer provider pairs (peers are ignored for the ranks)
        for i, rank in enumerate(self.as_graph.propagation_ranks):
            # Nothing to process at the start
            if i > 0:
                # Process first because maybe it recv from lower ranks
                for as_obj in rank:
                    as_obj.policy.process_incoming_anns(
                        from_rel=Relationships.CUSTOMERS,
                        propagation_round=propagation_round,
                        scenario=scenario,
                    )
            # Send to the higher ranks
            for as_obj in rank:
                as_obj.policy.propagate_to_providers()

    def _propagate_to_peers(self, propagation_round: int, scenario: "Scenario") -> None:
        """Propagate to peers"""

        # The reason you must separate this for loop here
        # is because propagation ranks do not take into account peering
        # It'd be impossible to take into account peering
        # since different customers peer to different ranks
        # So first do customer to provider propagation, then peer propagation
        for as_obj in self.as_graph:
            as_obj.policy.propagate_to_peers()
        for as_obj in self.as_graph:
            as_obj.policy.process_incoming_anns(
                from_rel=Relationships.PEERS,
                propagation_round=propagation_round,
                scenario=scenario,
            )

    def _propagate_to_customers(
        self, propagation_round: int, scenario: "Scenario"
    ) -> None:
        """Propagate to customers"""

        # Propogation ranks go from stubs to input_clique in ascending order
        # By customer provider pairs (peers are ignored for the ranks)
        # So here we start at the highest rank(input_clique) and propagate down
        for i, rank in enumerate(reversed(self.as_graph.propagation_ranks)):
            # There are no incomming Anns at the top
            if i > 0:
                for as_obj in rank:
                    as_obj.policy.process_incoming_anns(
                        from_rel=Relationships.PROVIDERS,
                        propagation_round=propagation_round,
                        scenario=scenario,
                    )
            for as_obj in rank:
                as_obj.policy.propagate_to_customers()

    ##############
    # JSON funcs #
    ##############

    def to_json(self) -> dict[str, Any]:
        """Returns a JSON representation of the simulation engine"""

        return {"as_graph": self.as_graph.to_json()}

    @classmethod
    def from_json(cls, json_obj: dict[str, Any]) -> "SimulationEngine":
        """Returns a SimulationEngine from a JSON representation"""

        return cls(as_graph=ASGraph.from_json(json_obj["as_graph"]))



from typing import TYPE_CHECKING, Iterator

from bgpsimulator.shared import PolicyPropagateInfo, Relationships, Settings, Timestamps
from bgpsimulator.simulation_engine import Announcement as Ann

from .rov import ROV

if TYPE_CHECKING:
    from bgpsimulator.as_graphs import AS
    from bgpsimulator.simulation_engine.policy.policy import Policy


class ROVPPV1Lite:
    """A Policy that deploys ROV++V1 Lite as defined in the ROV++ paper"""

    @staticmethod
    def get_policy_propagate_vals(
        policy: "Policy",
        neighbor_as: "AS",
        ann: "Ann",
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> PolicyPropagateInfo:
        if ann.rovpp_blackhole:
            return PolicyPropagateInfo(
                policy_propagate_bool=True, ann=ann, send_ann_bool=False
            )
        else:
            return PolicyPropagateInfo(
                policy_propagate_bool=False, ann=ann, send_ann_bool=True
            )

    @staticmethod
    def process_incoming_anns(
        policy: "Policy", from_rel: Relationships, propagation_round: int
    ) -> None:
        """Additional processing for incoming announcements"""

        ROVPPV1Lite.add_blackholes(policy, from_rel)
        ROVPPV1Lite.recount_blackholes(policy, propagation_round)

    @staticmethod
    def add_blackholes(policy: "Policy", from_rel: Relationships) -> None:
        """Adds blackholes announcements to the local RIB

        First add all non routed prefixes from ROAs as blackholes
        Then for each ann in th elocal RIB, if you recieived an
        invalid subprefix from the same neighbor,
        add it to the local RIB as a blackhole
        """

        non_routed_blackholes = ROVPPV1Lite.get_non_routed_blackholes_to_add(policy)
        routed_blackholes = ROVPPV1Lite.get_routed_blackholes_to_add(policy, from_rel)
        ROVPPV1Lite.add_blackholes_to_local_rib(
            policy, non_routed_blackholes + routed_blackholes
        )

    @staticmethod
    def get_non_routed_blackholes_to_add(policy: "Policy") -> list["Ann"]:
        """Gets all non routed blackholes to add to the local RIB"""

        non_routed_blackholes_to_add = []
        for roa in policy.route_validator.roas:
            if not roa.is_routed:
                non_routed_blackholes_to_add.append(
                    Ann(
                        prefix=roa.prefix,
                        next_hop_asn=policy.as_.asn,
                        as_path=(policy.as_.asn,),
                        # Victim's timestamp since it's upon ROA creation pre-attacker
                        timestamp=Timestamps.LEGITIMATE_ORIGIN,
                        recv_relationship=Relationships.ORIGIN,
                        rovpp_blackhole=True,
                    )
                )
        return non_routed_blackholes_to_add

    @staticmethod
    def get_routed_blackholes_to_add(
        policy: "Policy", from_rel: Relationships
    ) -> list["Ann"]:
        """Gets all routed blackholes from the anns you just recieved"""

        routed_blackholes_to_add = []
        for ann in policy.local_rib.values():
            for unprocessed_sub_ann in ROVPPV1Lite.invalid_subprefixes_from_neighbor(
                policy, ann
            ):
                processed_sub_ann = policy.process_ann(unprocessed_sub_ann, from_rel)
                # Add blackhole attributes to the processed ann
                blackhole_ann = processed_sub_ann.copy(
                    next_hop_asn=policy.as_.asn,
                    rovpp_blackhole=True,
                )
                routed_blackholes_to_add.append(blackhole_ann)
        return routed_blackholes_to_add

    @staticmethod
    def invalid_subprefixes_from_neighbor(
        policy: "Policy", ann: "Ann"
    ) -> Iterator["Ann"]:
        """Returns all invalid subprefixes from the neighbor"""

        # If we are the origin, then there are zero invalid anns from the same neighbor
        if ann.recv_relationship == Relationships.ORIGIN:
            return
        # for each subprefix ann recieved (NOTE: these aren't in local RIB since
        # they're invalid) and dropped by default (but they are recieved so we can
        # check there)
        for recvq_prefix, recvq_anns_list in policy.recv_q.items():
            # recvq prefix is a subprefix of ann
            if ann.prefix.supernet_of(recvq_prefix) and ann.prefix != recvq_prefix:
                for recvq_ann in recvq_anns_list:
                    if (
                        policy.ann_is_invalid_by_roa(recvq_ann)
                        # Check the first one in the path since it's already processed
                        and recvq_ann.as_path[0] == ann.as_path[1]
                    ):
                        yield recvq_ann

    @staticmethod
    def add_blackholes_to_local_rib(
        policy: "Policy", blackhole_anns: list["Ann"]
    ) -> None:
        """Adds blackholes to the local RIB"""

        for blackhole_ann in blackhole_anns:
            existing_ann = policy.local_rib.get(blackhole_ann.prefix)
            # Don't overwrite existing valid announcements
            if existing_ann is None:
                policy.local_rib[blackhole_ann.prefix] = blackhole_ann
            elif policy.ann_is_invalid_by_roa(existing_ann):
                # Not sure why anyone would ever need this
                # so I'm not implementing it for now
                # Also how would an existing ann be invalid in ROV++?
                # TODO: Implement this
                if policy.settings[Settings.BGP_FULL]:
                    raise NotImplementedError("Withdrawals not supported for ROV++")
                policy.local_rib[blackhole_ann.prefix] = blackhole_ann

    @staticmethod
    def recount_blackholes(policy: "Policy", propagation_round: int) -> None:
        """Recounts blackholes in the local RIB"""

        # It's possible that we had a previously valid prefix
        # Then later recieved a subprefix that was invalid
        # Or there was previously an invalid subprefix
        # But later that invalid subprefix was removed
        # So we must recount the holes of each ann in local RIB
        # NOTE June 22 2024: I think doing this may require the use of the AdjRIBsIn
        # because you could recieve a subprefix hijack round 1, and round 2 receive
        # the valid prefix from the same neighbor
        # not going to implement because of that, and because I don't think there's
        # a need to, because as far as I know there aren't any two round attacks
        # against ROV++. If someone comes up with one let me know and I can try to
        # help out, email at jfuruness@gmail.com.
        # NOTE: Additionally, we don't account for withdrawals at all...
        if propagation_round != 0:
            raise NotImplementedError("TODO: support ROV++ for multiple rounds")

    @staticmethod
    def valid_ann(policy: "Policy", ann: "Ann", from_rel: Relationships) -> bool:
        """Determines ROV++V1 Lite validity"""

        return ROV.valid_ann(policy, ann, from_rel)



from collections import defaultdict
from typing import TYPE_CHECKING, Any, cast
from weakref import proxy

from bgpsimulator.route_validator import RouteValidator
from bgpsimulator.shared import IPAddr, Prefix, Relationships, ROAValidity, Settings
from bgpsimulator.shared.exceptions import GaoRexfordError
from bgpsimulator.simulation_engine.announcement import Announcement as Ann

from .policy_extensions import (
    ASPA,
    ASPAPP,
    ASRA,
    BGP,
    ROST,
    ROV,
    AnnounceThenWithdraw,
    Leaker,
    ASPathEdgeFilter,
    ASPAwN,
    BGPiSecTransitive,
    BGPSec,
    EnforceFirstAS,
    FirstASNStrippingPrefixHijackCustomers,
    OnlyToCustomers,
    OriginPrefixHijackCustomers,
    PathEnd,
    PeerLockLite,
    PeerROV,
    ProviderConeID,
    RoSTTrustedRepository,
    ROVPPV1Lite,
    ROVPPV2iLite,
    ROVPPV2Lite,
    NeverPropagateWithdrawals,
)
from .adj_ribs_in import AdjRIBsIn
from .adj_ribs_out import AdjRIBsOut

if TYPE_CHECKING:
    from bgpsimulator.as_graphs import AS


class Policy:
    __slots__ = (
        "local_rib",
        "recv_q",
        "settings",
        "as_",
        "adj_ribs_in",
        "adj_ribs_out",
    )

    route_validator = RouteValidator()
    rost_trusted_repository = RoSTTrustedRepository()

    def __init__(
        self,
        as_: "AS",
        settings: tuple[bool] | None = None,
        local_rib: dict[Prefix, Ann] | None = None,
        adj_ribs_in: AdjRIBsIn | None = None,
        adj_ribs_out: AdjRIBsOut | None = None,
    ) -> None:
        """Add local rib and data structures here

        This way they can be easily cleared later without having to redo
        the graph

        This is also useful for regenerating an AS from YAML
        """

        self.local_rib: dict[Prefix, Ann] = local_rib or dict()
        self.recv_q: defaultdict[Prefix, list[Ann]] = defaultdict(list)
        self.adj_ribs_in: AdjRIBsIn = adj_ribs_in or AdjRIBsIn()
        self.adj_ribs_out: AdjRIBsOut = adj_ribs_out or AdjRIBsOut()
        self.rost_trusted_repository.clear()
        if settings:
            self.settings: tuple[bool, ...] = settings
        else:
            self.settings = tuple([False for _ in Settings])
        # The AS object that this routing policy is associated with
        # Casting this so we don't ened to put callable proxy type everywhere
        self.as_: AS = cast("AS", proxy(as_))

    def __eq__(self, other) -> bool:
        if isinstance(other, Policy):
            return self.to_json() == other.to_json()
        else:
            return NotImplemented

    def clear(self) -> None:
        """Clears the routing policy"""

        self.local_rib.clear()
        self.recv_q.clear()
        self.adj_ribs_in.clear()
        self.adj_ribs_out.clear()

    #########################
    # Process Incoming Anns #
    #########################

    def seed_ann(self, ann: Ann) -> None:
        """Seeds an announcement at this AS"""

        # Ensure we aren't replacing anything
        err = f"Seeding conflict {ann} {self.local_rib.get(ann.prefix)}"
        assert self.local_rib.get(ann.prefix) is None, err

        # If BGPSEC is deployed, modify the announcement
        if (
            self.settings[Settings.BGPSEC]
            or self.settings[Settings.BGP_I_SEC]
            or self.settings[Settings.BGP_I_SEC_TRANSITIVE]
        ):
            ann = BGPSec.get_modified_seed_ann(self, ann)
        # Seed by placing in the local rib
        self.local_rib[ann.prefix] = ann

    def receive_ann(self, ann: Ann) -> None:
        """Receives an announcement from a neighbor"""

        self.recv_q[ann.prefix].append(ann)

    def process_incoming_anns(
        self,
        *,
        from_rel: Relationships,
        propagation_round: int = 0,
        **kwargs,
    ) -> None:
        """Process all announcements that were incoming from a specific rel"""

        if self.settings[Settings.ROST]:
            ROST.preprocess_incoming_anns(
                self, from_rel=from_rel, propagation_round=propagation_round
            )

        # For each prefix, get all anns recieved
        for prefix, ann_list in self.recv_q.items():
            # Get announcement currently in local rib
            current_ann: Ann | None = self.local_rib.get(prefix)
            og_ann = current_ann

            # For each announcement that was incoming
            for new_ann in ann_list:
                # Ignore all withdrawals
                if self.settings[Settings.NEVER_WITHDRAW] and new_ann.withdraw:
                    continue

                if self.settings[Settings.BGP_FULL]:
                    # If withdrawal remove from AdjRIBsIn, otherwise add to AdjRIBsIn
                    self._process_new_ann_in_adj_ribs_in(new_ann, prefix, from_rel)

                # Process withdrawals even for invalid anns in the adj_ribs_in
                if new_ann.withdraw and self.settings[Settings.BGP_FULL]:
                    current_ann = self._remove_from_local_rib_and_get_new_best_ann(
                        new_ann, current_ann
                    )
                else:
                    # Get new best ann
                    current_ann = self._get_new_best_ann(current_ann, new_ann, from_rel)

            # This is a new best ann. Process it and add it to the local rib
            if og_ann != current_ann:
                if current_ann:
                    # Save to local rib
                    self.local_rib[current_ann.prefix] = current_ann
                if og_ann and self.settings[Settings.BGP_FULL]:
                    self.withdraw_ann_from_neighbors(
                        og_ann.copy(
                            next_hop_asn=self.as_.asn,
                            withdraw=True,
                        )
                    )

        # NOTE: all three of these have the same process_incoming_anns
        # which just adds ROV++ blackholes to the local RIB
        if (
            self.settings[Settings.ROVPP_V1_LITE]
            or self.settings[Settings.ROVPP_V2_LITE]
            or self.settings[Settings.ROVPP_V2I_LITE]
        ):
            ROVPPV1Lite.process_incoming_anns(self, from_rel, propagation_round)

        if self.settings[Settings.ROST]:
            ROST.postprocess_incoming_anns(self)

        self.recv_q.clear()

    def _get_new_best_ann(
        self, current_ann: Ann | None, new_ann: Ann, from_rel: Relationships
    ) -> Ann | None:
        """Checks new_ann's validity, processes it, returns best_ann_by_gao_rexford."""

        if self.valid_ann(new_ann, from_rel):
            new_ann_processed = self.process_ann(new_ann, from_rel)
            return self._get_best_ann_by_gao_rexford(current_ann, new_ann_processed)
        else:
            return current_ann

    def process_ann(self, unprocessed_ann: Ann, from_rel: Relationships) -> Ann:
        """Processes an announcement going from recv_q or adj_ribs_in to local rib.

        Must prepend yourself to the AS-path, change the recv_relationship, and add
        policy info if needed.
        """
        new_ann_processed = unprocessed_ann.copy(
            as_path=(self.as_.asn, *unprocessed_ann.as_path),
            recv_relationship=from_rel,
        )
        if (
            self.settings[Settings.BGP_I_SEC]
            or self.settings[Settings.BGP_I_SEC_TRANSITIVE]
        ):
            new_ann_processed = BGPiSecTransitive.process_ann(
                self, new_ann_processed, from_rel
            )
        elif self.settings[Settings.BGPSEC]:
            new_ann_processed = BGPSec.process_ann(self, new_ann_processed, from_rel)
        if self.settings[Settings.ROST]:
            new_ann_processed = ROST.process_ann(self, new_ann_processed, from_rel)
        return new_ann_processed

    def valid_ann(self, ann: Ann, from_rel: Relationships) -> bool:
        """Determine if an announcement is valid or should be dropped"""

        settings = self.settings

        if not BGP.valid_ann(self, ann, from_rel):
            return False
        # ASPAwN and ASRA are supersets of ASPA
        if (
            settings[Settings.ASPA]
            and not settings[Settings.ASRA]
            and not settings[Settings.ASPA_W_N]
            and not ASPA.valid_ann(self, ann, from_rel)
        ):
            return False
        if (
            settings[Settings.ASPA_W_N]
            and not settings[Settings.ASRA]
            and not ASPAwN.valid_ann(self, ann, from_rel)
        ):
            return False
        if settings[Settings.ASRA] and not ASRA.valid_ann(self, ann, from_rel):
            return False
        if settings[Settings.AS_PATH_EDGE_FILTER] and not ASPathEdgeFilter.valid_ann(
            self, ann, from_rel
        ):
            return False
        if settings[Settings.ENFORCE_FIRST_AS] and not EnforceFirstAS.valid_ann(
            self, ann, from_rel
        ):
            return False
        if settings[Settings.ONLY_TO_CUSTOMERS] and not OnlyToCustomers.valid_ann(
            self, ann, from_rel
        ):
            return False
        # All use ROV for validity
        if (
            settings[Settings.ROV]
            or settings[Settings.ROVPP_V1_LITE]
            or settings[Settings.ROVPP_V2_LITE]
            or settings[Settings.ROVPP_V2I_LITE]
        ) and not ROV.valid_ann(self, ann, from_rel):
            return False
        if settings[Settings.PEER_ROV] and not PeerROV.valid_ann(self, ann, from_rel):
            return False
        if settings[Settings.PATH_END] and not PathEnd.valid_ann(self, ann, from_rel):
            return False
        if settings[Settings.PEERLOCK_LITE] and not PeerLockLite.valid_ann(
            self, ann, from_rel
        ):
            return False
        if (
            settings[Settings.BGP_I_SEC] or settings[Settings.BGP_I_SEC_TRANSITIVE]
        ) and not BGPiSecTransitive.valid_ann(self, ann, from_rel):
            return False
        if settings[Settings.PROVIDER_CONE_ID] and not ProviderConeID.valid_ann(
            self, ann, from_rel
        ):
            return False
        if settings[Settings.ASPAPP] and not ASPAPP.valid_ann(self, ann, from_rel):
            return False

        return True

    def ann_is_invalid_by_roa(self, ann: Ann) -> bool:
        """Determines if an announcement is invalid by a ROA"""
        return ROAValidity.is_invalid(
            self.route_validator.get_roa_outcome(ann.prefix, ann.origin)[0]
        )

    ###############
    # Gao rexford #
    ###############

    def _get_best_ann_by_gao_rexford(
        self,
        current_ann: Ann | None,
        new_ann: Ann,
    ) -> Ann:
        """Determines if the new ann > current ann by Gao Rexford"""

        assert new_ann is not None, "New announcement can't be None"

        # When I had this as a list of funcs, it was 7x slower, resulting in bottlenecks
        # Gotta do it the ugly way unfortunately

        # mypy also doesn't understand that current_ann can not be None for these funcs
        final_ann = (
            (new_ann if current_ann is None else None)
            or self._get_best_ann_by_local_pref(current_ann, new_ann)  # type: ignore
            or self._get_best_ann_by_as_path(current_ann, new_ann)  # type: ignore
            # BGPSec is security third (see BGPSec class docstring)
            # NOTE: BGPiSec policies don't change path preference for easier deployment
            or (
                self.settings[Settings.BGPSEC]
                and BGPSec.get_best_ann_by_bgpsec(
                    self,
                    current_ann,  # type: ignore
                    new_ann,
                )
            )
            or self._get_best_ann_by_lowest_neighbor_asn_tiebreaker(
                current_ann,  # type: ignore
                new_ann,
            )
        )
        if final_ann:
            return final_ann
        else:
            raise GaoRexfordError("No ann was chosen")

    def _get_best_ann_by_local_pref(self, current_ann: Ann, new_ann: Ann) -> Ann | None:
        """Returns best announcement by local pref, or None if tie. Higher is better"""

        if current_ann.recv_relationship.value > new_ann.recv_relationship.value:
            return current_ann
        elif current_ann.recv_relationship.value < new_ann.recv_relationship.value:
            return new_ann
        # mypy requires this and it can't be turned off
        else:
            return None

    def _get_best_ann_by_as_path(self, current_ann: Ann, new_ann: Ann) -> Ann | None:
        """Returns best announcement by as path length, or None if tie.

        Shorter is better.
        """

        if len(current_ann.as_path) < len(new_ann.as_path):
            return current_ann
        elif len(current_ann.as_path) > len(new_ann.as_path):
            return new_ann
        # mypy requires this and it can't be turned off
        else:
            return None

    def _get_best_ann_by_lowest_neighbor_asn_tiebreaker(
        self, current_ann: Ann, new_ann: Ann
    ) -> Ann:
        """Determines if the new ann > current ann by Gao Rexford for ties

        This breaks ties by lowest asn of the neighbor sending the announcement
        So if the two announcements are from the same neighbor, return current ann
        """

        current_neighbor_asn = current_ann.as_path[min(len(current_ann.as_path), 1)]
        new_neighbor_asn = new_ann.as_path[min(len(new_ann.as_path), 1)]

        return current_ann if current_neighbor_asn <= new_neighbor_asn else new_ann

    def pre_propagation_hook(
        self, propagation_round: int, scenario: "Scenario"
    ) -> None:
        """Before propagating to anyone, pre-propagation hook"""

        # Withdraws announcements round 2
        if self.settings[Settings.ANNOUNCE_THEN_WITHDRAW]:
            AnnounceThenWithdraw.pre_propagation_hook(self, propagation_round, scenario)
        if self.settings[Settings.LEAKER]:
            Leaker.pre_propagation_hook(self, propagation_round, scenario)

    def propagate_to_customers(self) -> None:
        """Propogates to customers anns that have a known recv_rel"""

        send_rels: set[Relationships] = {
            Relationships.ORIGIN,
            Relationships.CUSTOMERS,
            Relationships.PEERS,
            Relationships.PROVIDERS,
        }
        self._propagate(Relationships.CUSTOMERS, send_rels)

    def propagate_to_peers(self) -> None:
        """Propogates to peers anns from this AS (origin) or from customers"""

        send_rels: set[Relationships] = {Relationships.ORIGIN, Relationships.CUSTOMERS}
        self._propagate(Relationships.PEERS, send_rels)

    def propagate_to_providers(self) -> None:
        """Propogates to providers anns that have recv_rel from origin or customers"""

        send_rels: set[Relationships] = {Relationships.ORIGIN, Relationships.CUSTOMERS}
        self._propagate(Relationships.PROVIDERS, send_rels)

    def _propagate(
        self, propagate_to: Relationships, send_rels: set[Relationships]
    ) -> None:
        """Propogates announcements from local rib to other ASes

        send_rels are the relationships that are acceptable to send
        """

        neighbor_ases = self.as_.get_neighbor(propagate_to)

        for _prefix, unprocessed_ann in self.local_rib.items():
            # We must set the next_hop when sending
            # Copying announcements is a bottleneck for sims,
            # so we try to do this as little as possible
            if neighbor_ases and unprocessed_ann.recv_relationship in send_rels:
                ann = unprocessed_ann.copy(next_hop_asn=self.as_.asn)
            else:
                continue

            for neighbor_as in neighbor_ases:
                if ann.recv_relationship in send_rels and (
                    not self.settings[Settings.BGP_FULL]
                    or not self._prev_sent(neighbor_as, ann)
                ):
                    # Policy took care of it's own propagation for this ann
                    if self.policy_propagate(neighbor_as, ann, propagate_to, send_rels):
                        continue
                    else:
                        self.process_outgoing_ann(
                            neighbor_as, ann, propagate_to, send_rels
                        )

    def policy_propagate(
        self,
        neighbor_as: "AS",
        ann: Ann,
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> bool:
        """Policies can override this to handle their own propagation and return True

        This can no longer be as simple as it was in BGPy since many policies may
        interact with one another. So there are three values returned from each policy:
        1. policy_propagate_bool. If this is False, the policy did nothing, if it is
        true, Policy should return True from this method
        2. ann: The modified ann. This can be then be passed into the other funcs
        3. send_ann_bool. Sometimes a policy may declare the ann shoulldn't be sent at
        all (for example, ROV++V1 won't send blackholes), in which case, just return
        True immediately without sending any anns
        """

        og_ann = ann
        if (
            self.settings[Settings.BGP_I_SEC]
            or self.settings[Settings.BGP_I_SEC_TRANSITIVE]
        ):
            policy_propagate_info = BGPiSecTransitive.get_policy_propagate_vals(
                self, neighbor_as, ann, propagate_to, send_rels
            )
            if policy_propagate_info.policy_propagate_bool:
                ann = policy_propagate_info.ann
                if not policy_propagate_info.send_ann_bool:
                    return True
        # NOTE: THIS MUST BE ELIF!! BGPiSecTransitive is a superset of BGPSec and has
        # different get_policy_propagate_vals
        elif self.settings[Settings.BGPSEC]:
            policy_propagate_info = BGPSec.get_policy_propagate_vals(
                self, neighbor_as, ann, propagate_to, send_rels
            )
            if policy_propagate_info.policy_propagate_bool:
                ann = policy_propagate_info.ann
                if not policy_propagate_info.send_ann_bool:
                    return True

        if self.settings[Settings.ONLY_TO_CUSTOMERS]:
            policy_propagate_info = OnlyToCustomers.get_policy_propagate_vals(
                self, neighbor_as, ann, propagate_to, send_rels
            )
            if policy_propagate_info.policy_propagate_bool:
                ann = policy_propagate_info.ann
                if not policy_propagate_info.send_ann_bool:
                    return True

        if self.settings[Settings.ROVPP_V2I_LITE]:
            policy_propagate_info = ROVPPV2iLite.get_policy_propagate_vals(
                self, neighbor_as, ann, propagate_to, send_rels
            )
            if policy_propagate_info.policy_propagate_bool:
                ann = policy_propagate_info.ann
                if not policy_propagate_info.send_ann_bool:
                    return True
        # If V2i is deployed, don't use V2
        elif self.settings[Settings.ROVPP_V2_LITE]:
            policy_propagate_info = ROVPPV2Lite.get_policy_propagate_vals(
                self, neighbor_as, ann, propagate_to, send_rels
            )
            if policy_propagate_info.policy_propagate_bool:
                ann = policy_propagate_info.ann
                if not policy_propagate_info.send_ann_bool:
                    return True
        # If v2i or v2 are set, don't use v1 (since they are supersets)
        elif self.settings[Settings.ROVPP_V1_LITE]:
            policy_propagate_info = ROVPPV1Lite.get_policy_propagate_vals(
                self, neighbor_as, ann, propagate_to, send_rels
            )
            if policy_propagate_info.policy_propagate_bool:
                ann = policy_propagate_info.ann
                if not policy_propagate_info.send_ann_bool:
                    return True

        if self.settings[Settings.ORIGIN_PREFIX_HIJACK_CUSTOMERS]:
            policy_propagate_info = (
                OriginPrefixHijackCustomers.get_policy_propagate_vals(
                    self, neighbor_as, ann, propagate_to, send_rels
                )
            )
            if policy_propagate_info.policy_propagate_bool:
                ann = policy_propagate_info.ann
                if not policy_propagate_info.send_ann_bool:
                    return True
        if self.settings[Settings.FIRST_ASN_STRIPPING_PREFIX_HIJACK_CUSTOMERS]:
            policy_propagate_info = (
                FirstASNStrippingPrefixHijackCustomers.get_policy_propagate_vals(
                    self, neighbor_as, ann, propagate_to, send_rels
                )
            )
            if policy_propagate_info.policy_propagate_bool:
                ann = policy_propagate_info.ann
                if not policy_propagate_info.send_ann_bool:
                    return True

        if self.settings[Settings.NEVER_PROPAGATE_WITHDRAWALS]:
            policy_propagate_info = NeverPropagateWithdrawals.get_policy_propagate_vals(
                self, neighbor_as, ann, propagate_to, send_rels
            )
            if policy_propagate_info.policy_propagate_bool:
                ann = policy_propagate_info.ann
                if not policy_propagate_info.send_ann_bool:
                    return True

        if og_ann != ann:
            if not ann.withdraw and self.settings[Settings.BGP_FULL]:
                self.adj_ribs_out.add_ann(neighbor_as.asn, ann)
            self.process_outgoing_ann(neighbor_as, ann, propagate_to, send_rels)
            return True
        else:
            return False

    def process_outgoing_ann(
        self,
        neighbor_as: "AS",
        ann: Ann,
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> None:
        """Adds ann to the neighbors recv q"""

        if not ann.withdraw and self.settings[Settings.BGP_FULL]:
            self.adj_ribs_out.add_ann(neighbor_as.asn, ann)
        # Add the new ann to the incoming anns for that prefix
        neighbor_as.policy.receive_ann(ann)

    #########################
    # Data Plane Validation #
    #########################

    def get_most_specific_ann(self, dest_ip_addr: IPAddr) -> Ann | None:
        """Returns the most specific announcement for a destination IP address

        Uses caching whenever possible to avoid expensive lookups at each AS
        however, don't cache large RIBs, there won't be duplicates,
        and don't keep too many in the cache, there won't be duplicates
        We need to watch our RAM here

        NOTE: Caching actually slowed this down by about 1.5x so we don't do it anymore
        """

        matching_prefixes = sorted(
            (p for p in self.local_rib if p.supernet_of(dest_ip_addr)),
            key=lambda p: p.prefixlen,
            reverse=True,
        )
        most_specific_prefix = matching_prefixes[0] if matching_prefixes else None

        return self.local_rib[most_specific_prefix] if most_specific_prefix else None

    def passes_sav(self, dest_ip_addr: IPAddr, most_specific_ann: Ann) -> bool:
        """Determines if the AS passes the source address validation check"""

        return True

    ##################################################
    # BGPFull (withdrawals, ribs in, ribs out) funcs #
    ##################################################

    def _process_new_ann_in_adj_ribs_in(
        self, unprocessed_ann: Ann, prefix: Prefix, from_rel: Relationships
    ) -> None:
        """Adds ann to ribs in if the ann is not a withdrawal"""

        # Remove ann using withdrawal from AdjRIBsIn
        if unprocessed_ann.withdraw:
            neighbor = unprocessed_ann.as_path[0]
            # Remove ann from Ribs in
            self.adj_ribs_in.remove_entry(neighbor, prefix)
        # Add ann to AdjRIBsIn
        else:
            self.adj_ribs_in.add_unprocessed_ann(unprocessed_ann, from_rel)

    def _remove_from_local_rib_and_get_new_best_ann(
        self, new_ann: "Ann", local_rib_ann: "Ann | None"
    ) -> "Ann | None":
        # This is for removing the original local RIB ann
        if (
            local_rib_ann
            and new_ann.prefix == local_rib_ann.prefix
            # new_ann is unproccessed
            and new_ann.as_path == local_rib_ann.as_path[1:]
            and local_rib_ann.recv_relationship != Relationships.ORIGIN
        ):
            # Withdrawal exists in the local RIB, so remove it and reset current ann
            self.local_rib.pop(new_ann.prefix, None)
            local_rib_ann = None
            # Get the new best ann thus far
            processed_best_adj_ribs_in_ann = self._get_and_process_best_adj_ribs_in_ann(
                new_ann.prefix
            )
            if processed_best_adj_ribs_in_ann:
                local_rib_ann = self._get_best_ann_by_gao_rexford(
                    local_rib_ann,
                    processed_best_adj_ribs_in_ann,
                )

        return local_rib_ann

    def withdraw_ann_from_neighbors(self, withdraw_ann: Ann) -> None:
        """Withdraw a route from all neighbors.

        This function will not remove an announcement from the local rib, that
        should be done before calling this function.

        Note that withdraw_ann is a deep copied ann
        """
        assert withdraw_ann.withdraw is True
        assert withdraw_ann.next_hop_asn == self.as_.asn
        if self.settings[Settings.ROST]:
            ROST.withdraw_ann_from_neighbors(self, withdraw_ann)
        # Check adj_ribs_out to see where the withdrawn ann was sent
        for send_neighbor_asn in self.adj_ribs_out.populated_neighbors():
            # Delete ann from ribs out
            removed = self.adj_ribs_out.remove_entry(
                send_neighbor_asn, withdraw_ann.prefix
            )
            # If the announcement was sent to that neighbor
            if removed:
                send_rels = set(Relationships)
                if send_neighbor_asn in self.as_.customer_asns:
                    propagate_to = Relationships.CUSTOMERS
                elif send_neighbor_asn in self.as_.provider_asns:
                    propagate_to = Relationships.PROVIDERS
                elif send_neighbor_asn in self.as_.peer_asns:
                    propagate_to = Relationships.PEERS
                else:
                    raise NotImplementedError("Case not accounted for")
                send_neighbor = self.as_.as_graph.as_dict[send_neighbor_asn]
                # Policy took care of it's own propagation for this ann
                if self.policy_propagate(
                    send_neighbor, withdraw_ann, propagate_to, send_rels
                ):
                    continue
                else:
                    self.process_outgoing_ann(
                        send_neighbor, withdraw_ann, propagate_to, send_rels
                    )

    def _get_and_process_best_adj_ribs_in_ann(self, prefix: Prefix) -> "Ann | None":
        """Selects best ann from ribs in (remember, AdjRIBsIn is unprocessed"""

        # Get the best announcement
        best_ann: Ann | None = None
        for ann_info in self.adj_ribs_in.get_ann_infos(prefix):
            # This also processes the announcement
            best_ann = self._get_new_best_ann(
                best_ann, ann_info.unprocessed_ann, ann_info.recv_relationship
            )
        return best_ann

    def _prev_sent(self, neighbor: "AS", ann: Ann) -> bool:
        """Don't send what we've already sent"""

        return ann == self.adj_ribs_out.get_ann(neighbor.asn, ann.prefix)

    ##############
    # JSON funcs #
    ##############

    def to_json(self) -> dict[str, Any]:
        """Converts the routing policy to a JSON object"""
        return {
            "local_rib": {
                str(prefix): ann.to_json() for prefix, ann in self.local_rib.items()
            },
            "settings": list(self.settings),
            "adj_ribs_in": self.adj_ribs_in.to_json(),
            "adj_ribs_out": self.adj_ribs_out.to_json(),
        }

    @classmethod
    def from_json(cls, json_obj: dict[str, Any], as_: "AS") -> "Policy":
        return cls(
            as_=as_,
            local_rib={
                Prefix(prefix): Ann.from_json(ann)
                for prefix, ann in json_obj.get("local_rib", {}).items()
            },
            adj_ribs_in=AdjRIBsIn.from_json(json_obj.get("adj_ribs_in", {})),
            adj_ribs_out=AdjRIBsOut.from_json(json_obj.get("adj_ribs_out", {})),
            settings=tuple(json_obj.get("settings", [])),
        )



from bgpsimulator.route_validator import ROA
from bgpsimulator.shared import IPAddr
from bgpsimulator.shared.enums import CommonPrefixes, Relationships, Timestamps
from bgpsimulator.simulation_engine import Announcement as Ann
from bgpsimulator.simulation_engine import SimulationEngine
from bgpsimulator.simulation_framework.scenarios.scenario import Scenario


class PrefixHijack(Scenario):
    """Victim announces a prefix covered by a ROA, attacker announces the same
    prefix (invalid by ROA).
    """

    def _get_seed_asn_ann_dict(self, engine: SimulationEngine) -> dict[int, list[Ann]]:
        anns = dict()
        for legitimate_origin_asn in self.legitimate_origin_asns:
            anns[legitimate_origin_asn] = [
                Ann(
                    prefix=CommonPrefixes.PREFIX.value,
                    as_path=(legitimate_origin_asn,),
                    next_hop_asn=legitimate_origin_asn,
                    recv_relationship=Relationships.ORIGIN,
                    timestamp=Timestamps.LEGITIMATE_ORIGIN,
                )
            ]

        for attacker_asn in self.attacker_asns:
            anns[attacker_asn] = [
                Ann(
                    prefix=CommonPrefixes.PREFIX.value,
                    as_path=(attacker_asn,),
                    next_hop_asn=attacker_asn,
                    recv_relationship=Relationships.ORIGIN,
                    timestamp=Timestamps.ATTACKER,
                )
            ]

        return anns

    def _get_roas(
        self,
        *,
        seed_asn_ann_dict: dict[int, list[Ann]],
        engine: SimulationEngine,
    ) -> list[ROA]:
        """Returns a list of ROAs"""

        return [
            ROA(CommonPrefixes.PREFIX.value, x) for x in self.legitimate_origin_asns
        ]

    def _get_dest_ip_addr(self) -> IPAddr:
        """Returns the destination IP address for the scenario"""

        return IPAddr("1.2.3.4")



from .as_graph import ASGraph, ASGraphUtils, AS
from .as_graph_generators import CAIDAASGraphCollector, CAIDAASGraphJSONConverter

__all__ = [
    "ASGraph",
    "ASGraphUtils",
    "AS",
    "CAIDAASGraphCollector",
    "CAIDAASGraphJSONConverter",
]



from pathlib import Path

from bgpsimulator.shared import Settings
from bgpsimulator.simulation_framework import (
    ScenarioConfig,
    Simulation,
    SubprefixHijack,
)


def main():
    """Runs the defaults"""

    sim = Simulation(
        percent_ases_randomly_adopting=(
            10,
            20,
            50,
            80,
            99,
        ),
        scenario_configs=(
            ScenarioConfig(
                label="Subprefix Hijack; ROV Adopting",
                ScenarioCls=SubprefixHijack,
                default_adoption_settings={
                    Settings.ROV: True,
                },
            ),
        ),
        output_dir=Path("~/Desktop/sims/main_ex").expanduser(),
    )
    sim.run()


if __name__ == "__main__":
    main()



class CycleError(RuntimeError):
    pass


class NoCAIDAURLError(Exception):
    """Raised when no CAIDA URL is found"""

    pass


class GaoRexfordError(RuntimeError):
    """Error that occurs during gao rexford, such as failing to choose an ann"""

    pass


class AnnouncementNotFoundError(RuntimeError):
    """Exception that covers when an Announcement isn't findable

    ex: in local RIB, in AdjRIBsIn, etc
    """

    pass


class ReservedPrefixError(ValueError):
    """Exception that covers when a prefix is reserved"""

    pass


class InvalidIPAddressError(ValueError):
    """Exception that covers when an IP address is invalid"""

    pass



from .constants import SINGLE_DAY_CACHE_DIR, bgpsimulator_logger
from .exceptions import (
    CycleError,
    NoCAIDAURLError,
    GaoRexfordError,
    AnnouncementNotFoundError,
    ReservedPrefixError,
    InvalidIPAddressError,
)
from .enums import (
    Relationships,
    Settings,
    ROAValidity,
    ROARouted,
    ASNGroups,
    InAdoptingASNs,
    Outcomes,
    CommonPrefixes,
    Timestamps,
    CommonASNs,
)
from .ip_addr import IPAddr
from .prefix import Prefix
from .policy_propagate_info import PolicyPropagateInfo

__all__ = [
    "SINGLE_DAY_CACHE_DIR",
    "bgpsimulator_logger",
    "CycleError",
    "NoCAIDAURLError",
    "GaoRexfordError",
    "AnnouncementNotFoundError",
    "ReservedPrefixError",
    "InvalidIPAddressError",
    "Relationships",
    "Settings",
    "Prefix",
    "ROAValidity",
    "ROARouted",
    "IPAddr",
    "ASNGroups",
    "InAdoptingASNs",
    "Outcomes",
    "CommonPrefixes",
    "Timestamps",
    "PolicyPropagateInfo",
    "CommonASNs",
]
